import json
import os
import pandas as pd
from pandas import DataFrame
import datetime
from sensiml.datamanager.featurefiles import FeatureFiles
from sensiml.datamanager.captures import Captures
from sensiml.datamanager.knowledgepack import KnowledgePack
from sensiml.datamanager.sandboxes import Sandboxes
from sensiml.datamanager.queries import Queries
from sensiml.method_calls.functioncall import FunctionCall
from sensiml.datamanager.capture_configurations import CaptureConfigurations
import sensiml.base.utility as utility


class Project(object):
    """Base class for a project."""

    _uuid = ""
    _name = ""
    _created_at = None
    _schema = {}
    _settings = {}
    _query_optimized = True

    def __init__(self, connection):
        """Initialize a project instance.

            Args:
                connection (connection object)
        """
        self._connection = connection
        self._feature_files = FeatureFiles(self._connection, self)
        self._captures = Captures(self._connection, self)
        self._sandboxes = Sandboxes(self._connection, self)
        self._queries = Queries(self._connection, self)
        self._plugin_config = None
        self._capture_configurations = CaptureConfigurations(
            self._connection, self)

    @property
    def uuid(self):
        """Auto generated unique identifier for the project object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """Name of the project object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def created_at(self):
        """Date of the project creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        if value:
            self._created_at = datetime.datetime.strptime(
                value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
            )

    @property
    def schema(self):
        """Schema of the project object"""
        return self._schema

    @schema.setter
    def schema(self, value):
        self._schema = value

    @property
    def plugin_config(self):
        """Plugin Config of the project object"""
        return self._plugin_config

    @plugin_config.setter
    def plugin_config(self, value):
        self._plugin_config = value

    @property
    def settings(self):
        """Global settings of the project object"""
        return self._settings

    @property
    def query_optimized(self):
        return self._query_optimized

    def add_segmenter(self, name, segmenter, preprocess=None, custom=False):
        """Saves a segmentation algorithm as the project's global segmentation setting.

            Args:
                name(str): Name to call the segmenter
                segmenter(FunctionCall): segmentation call object that the project will use by default
                preprocess(dict): Segment transforms to perform before segmenter
                custom(bool): a custom segmenter, or one of our server side segmenters


        """
        url = "project/{0}/segmenter/".format(self.uuid)
        if segmenter is not None:
            if not isinstance(segmenter, FunctionCall):
                print("segmenter is not a function call.")
                return
            segmenter_dict = segmenter._to_dict()
            if not segmenter_dict["type"] == "segmenter":
                print("segmenter is not a function call for a segmenter")
                return
            parameters = json.dumps(segmenter_dict)
        else:
            parameters = None

        if not isinstance(custom, bool):
            print("Custom must either be true or false.")
            return

        if preprocess:
            if isinstance(preprocess, dict):
                preprocess = json.dumps(preprocess)

        segmenter_info = {
            "custom": custom,
            "name": str(name),
            "parameters": parameters,
            "preprocess": preprocess,
        }

        request = self._connection.request("post", url, segmenter_info)
        response, err = utility.check_server_response(request)
        if err is False:
            print("Segmenter Uploaded.")
            return response

    def insert(self):
        """Calls the REST API to insert a new object, uses only the name and schema."""
        url = "project/"
        project_info = {
            "name": self.name,
            "capture_sample_schema": self.schema,
            "settings": self.settings,
            "plugin_config": self.plugin_config,
        }
        request = self._connection.request("post", url, project_info)
        response, err = utility.check_server_response(request)
        if err is False:
            self.uuid = response["uuid"]
            self._settings = response["settings"]
            self._query_optimized = response.get("optimized", True)
            self.plugin_config = response["plugin_config"]

    def update(self):
        """Calls the REST API to update the object."""
        url = "project/{0}/".format(self.uuid)
        project_info = {
            "name": self.name,
            "capture_sample_schema": self.schema,
            "settings": self.settings,
            "plugin_config": self.plugin_config,
        }
        request = self._connection.request("patch", url, project_info)
        response, err = utility.check_server_response(request)
        if err is False:
            self.plugin_config = response.get("plugin_config", None)

    def delete(self):
        """Calls the REST API to delete the object."""
        url = "project/{0}/".format(self.uuid)
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request)

    def refresh(self):
        """Calls the REST API and self populates from the server."""
        url = "project/{0}/".format(self.uuid)
        request = self._connection.request("get", url)
        response, err = utility.check_server_response(request)
        if err is False:
            self.name = response["name"]
            self.schema = response["capture_sample_schema"]
            self._query_optimized = response.get("optimized", True)
            self.plugin_config = response.get("plugin_config", None)
            self.created_at = response.get("created_at", None)

    def query_optimize(self, force=False):
        self.refresh()
        """Calls the REST API and optimizes or re-optimizes the project for querying."""
        if not self.schema or not len(self._captures.get_captures()):
            print(
                "Cannot query optimize {} until there are uploaded captures.".format(
                    self._name
                )
                + "If data was uploaded, try dsk.project.refresh() followed by dsk.project.query_optimize()."
            )
        elif not self._query_optimized:
            print(
                "{} is not optimized for querying. Optimizing now...".format(
                    self._name)
            )
            self._create_profile()
        elif force:
            print("Re-optimizing {} for querying now...".format(self._name))
            self._delete_profile()
            self._create_profile()
        else:
            print("Database already optimized for query.")

    def _create_profile(self):
        """Calls the REST API and creates a profile for optimized query times."""
        url = "project/{0}/profile/".format(self.uuid)
        request = self._connection.request("post", url)
        response, err = utility.check_server_response(request, is_octet=True)
        self.refresh()

    def _delete_profile(self):
        """Calls the REST API and drops the project profile."""
        url = "project/{0}/profile/".format(self.uuid)
        # Make a call to delete the profile
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request, is_octet=True)
        self.refresh()

    def get_knowledgepack(self, kp_uuid):
        """Gets the KnowledgePack(s) created by the sandbox.

            Returns:
                a KnowledgePack instance, list of instances, or None
        """
        url = "knowledgepack/{0}/".format(kp_uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            kp = KnowledgePack(
                self._connection, self.uuid, response_data.get("sandbox_uuid")
            )
            kp.initialize_from_dict(response_data)
            return kp

    def _get_knowledgepacks(self):
        """Gets the KnowledgePack(s) created by the sandbox.

            Returns:
                a KnowledgePack instance, list of instances, or None
        """
        url = "project/{0}/knowledgepack/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return DataFrame(response_data)

    def list_knowledgepacks(self):
        """Lists all of the knowledgepacks associated with this project

        Returns:
            DataFrame: knowledpacks on kb cloud
        """

        knowledgepacks = self._get_knowledgepacks().rename(
            columns={
                "name": "Name",
                "project_name": "Project",
                "sandbox_name": "Pipeline",
                "uuid": "kp_uuid",
                "created_at": "Created",
                "knowledgepack_description": "kp_description",
            }
        )

        if len(knowledgepacks) < 1:
            print("No Knowledgepacks stored for this project on the cloud.")
            return None
        return knowledgepacks[knowledgepacks["Name"] != ""][
            ["Name", "Created", "Project", "Pipeline", "kp_uuid", "kp_description"]
        ]

    def initialize_from_dict(self, dict):
        """Reads a json dict and populates a single project.

            Args:
                dict (dict): contains the project's 'name', 'uuid', 'schema', and 'settings' properties
        """
        self.uuid = dict["uuid"]
        self.name = dict["name"]
        self.schema = dict["capture_sample_schema"]
        self._settings = dict.get("settings", [])
        self._query_optimized = dict.get("optimized", True)
        self.plugin_config = dict.get("plugin_config", None)
        self.created_at = dict.get("created_at", None)

    def __getitem__(self, key):
        if type(key) is str:
            return self.captures.get_capture_by_filename(key)
        else:
            return self.captures.get_captures()[key]

    @property
    def featurefiles(self):
        return self._feature_files

    @property
    def captures(self):
        return self._captures

    @property
    def sandboxes(self):
        return self._sandboxes

    @property
    def queries(self):
        return self._queries

    @property
    def capture_configurations(self):
        return self._capture_configurations

    def columns(self):
        """Returns the sensor columns available in the project.

            Returns:
                columns (list[str]): a list of string names of the project's sensor columns
        """
        try:
            columnnames = self.schema.keys()
            return columnnames
        except:
            return None

    def metadata_columns(self):
        """Returns the metadata columns available in the project.

            Returns:
                columns (list[str]): a list of string names of the project's metadata columns
        """
        return self.captures.get_metadata_names()

    def metadata_values(self):
        return self.captures.get_metadata_names_and_values()

    def label_values(self):
        return self.captures.get_label_names_and_values()

    def label_columns(self):
        """Returns the label columns available in the project.

            Returns:
                columns (list[str]): a list of string names of the project's metadata columns
        """
        return self.captures.get_label_names()

    def statistics(self):
        """Gets all capture statistics for the project.

            Returns:
                DataFrame of capture statistics
        """
        url = "project/{0}/statistics/".format(self.uuid)
        data = {"events": False}
        response = self._connection.request("get", url, json=data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            data, error_report = utility.make_statistics_table(response_data)
            return data, error_report
        else:
            return None

    def get_active_pipelines(self):
        url = "project/{0}/active/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            if response_data:
                return response_data

        return None

    def get_segmenters(self):
        url = "project/{0}/segmenter/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            if response_data:
                return DataFrame(response_data).set_index("id")

        return None

    def convert_project_to_csv(
        self, dclprojpath, name=None, label_name=None, segmenter_id=None
    ):
        """
        Args:
            dclprojpath: path to dcl project folder
            name (str): name of the dcl project folder. Defaults to current projects name.
            label_name: name of the label column. Defaults to current projects label column.
            segment_id: id of the segmenter folder to use.
        """

        if name is None:
            name = self.name

        dcl_project_name = os.path.join(dclprojpath, name)

        if not os.path.exists(dcl_project_name):
            print("Error: DCL project does not exist at " + dcl_project_name)
            return

        segmenters = None
        if segmenter_id is None:
            segmenters = self.get_segmenters()
            segmenter_id = int(
                input(
                    "Select Segementer ID\n\n"
                    + "\n".join(
                        [
                            "\t name: {}: id {}".format(x[0], x[1])
                            for x in zip(segmenters[["name"]].values, segmenters.index)
                        ]
                    )
                    + "\n"
                )
            )

        if segmenter_id:
            if not os.path.exists(
                os.path.join(dcl_project_name, "segments", str(segmenter_id))
            ):
                print("Invalid Segmenter ID")
                return

        metadata_folder = os.path.join(dcl_project_name, "metadata")
        segments_folder = os.path.join(
            dcl_project_name, "segments", str(segmenter_id))
        data_folder = os.path.join(dcl_project_name, "data")

        label_config = json.load(
            open(os.path.join(dcl_project_name, "Label.config")))

        label_name = label_config["task_list"][0]["LabelName"]

        root_names = [x[:-4]
                      for x in os.listdir(data_folder) if x[-4:] == ".csv"]

        M = []
        for root_name in root_names:
            print("Reading File:", root_name)

            if not os.path.exists(os.path.join(segments_folder, root_name + ".sdcl")):
                print("No Segments for this Segmenter, Skipping.")
                continue

            tmp_df = pd.read_csv(os.path.join(data_folder, root_name + ".csv"))

            print("\tInitial Row #:", tmp_df.shape[0])

            # Set this to the Label you want to look at
            tmp_df[label_name] = None
            tmp_df["SegmentID"] = None
            tmp_df["filename"] = root_name

            segments = json.load(
                open(os.path.join(segments_folder, root_name + ".sdcl"))
            )
            for segment in segments["DetectedSegments"]:
                for label_data in segment["LabelData"]:
                    tmp_df.loc[
                        label_data["capture_sample_sequence_start"]: label_data[
                            "capture_sample_sequence_end"
                        ],
                        label_data["name"],
                    ] = label_data["value"]

            tmp_df.dropna(subset=[label_name], inplace=True)

            if os.path.exists(os.path.join(metadata_folder, root_name + ".dcl")):
                metadata = json.load(
                    open(os.path.join(metadata_folder, root_name + ".dcl"))
                )
                for meta in metadata["FileMetadata"]:
                    tmp_df[meta["name"]] = meta["value"]

            print("\tSegments #:", len(tmp_df.SegmentID.unique()))
            print("\tFinal Row #:", tmp_df.shape[0])
            M.append(tmp_df)

        return pd.concat(M, sort=False).reset_index(drop=True)



import os
import json
import sensiml.base.utility as utility


class Segmetner(object):
    """Base class for a segmenter object."""

    def __init__(self, connection, project):
        """Initialize a metadata object.

            Args:
                connection
                project
        """
        self._uuid = ""
        self._name = ""
        self._parameters = None
        self._preprocess = None
        self._custom = True
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    @property
    def preprocess(self):
        return self._preprocess

    @preprocess.setter
    def preprocess(self, value):
        self._preprocess = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/segmenter/".format(
            self._project.uuid
        )
        label_info = {
            "name": self.name,
            "preprocess": self.preprocess,
            "parameters": self.parameters,
        }
        response = self._connection.request("post", url, label_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["id"]

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/segmenter/{1}/".format(
            self._project.uuid, self.uuid
        )
        label_info = {
            "name": self.name,
            "preprocess": self.preprocess,
            "parameters": self.parameters,
        }
        response = self._connection.request("put", url, label_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/segmenter/{1}/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/segmenter/{1}/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                dict (dict): contains the uuid, name, type
        """
        self.uuid = data["id"]
        self.name = data["name"]
        self.preprocess = data['preporcess']
        self.parameters = data['parameters']


class SegmenterSet(object):
    """Base class for a segmenter object."""

    def __init__(self, connection, project, initialize_set=True):
        self._connection = connection
        self._project = project
        self._set = None

        if initialize_set:
            self.refresh()

    @property
    def segmenters(self):
        if self._set is None:
            self._set = self.get_set()

        return self._set

    def refresh(self):
        self._set = self.get_set()


    def get_set(self):
        """Gets a list of all segmenters in the project.

        Returns:
            list (segmenters)
        """
        err = False
        url = "project/{0}/segmenter/".format(self._project.uuid)
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)
        # Populate the retrieved featurefiles
        segmenters = []

        if err is False:
            try:
                for segmenter_params in response_data:
                    segmenters.append(segmenter_params)
            except:
                pass

        return segmenters



import os
import zipfile
from os import listdir, makedirs
from os.path import isfile, join, exists, splitext
from pathlib import Path

from ipywidgets import Layout, VBox, HBox, Button
from ipywidgets import widgets

from sensiml.device_flashing import *
from sensiml.widgets.base_widget import BaseWidget


class FlashWidget(BaseWidget):
    def __init__(self, dsk=None, folder="knowledgepacks"):
        self._dsk = dsk
        self._folder = Path(folder)
        self._flashing_file = ""
        self._selected_flasher = None
        if not exists(folder):
            makedirs(folder)

    def _refresh_platform_files(self, b=None):
        key_name = self.platform_widget.value.replace(" ", "-")
        platform_search = "{}".format(key_name).lower()
        ext_match = ".zip"
        self.flashable_files_widget.options = []
        platform_files = [
            f
            for f in listdir(self._folder)
            if isfile(join(self._folder, f))
            and platform_search in f.lower()
            and splitext(f)[-1] == ext_match
            and "bin" in f.lower()
        ]

        if len(platform_files) > 0:
            self.flashable_files_widget.options = platform_files
            self.flashable_files_widget.value = platform_files[0]

    def _refresh_flash_methods(self):
        if self._selected_flasher is None:
            return
        self.flash_methods_widget.options = self._selected_flasher.flash_methods

    def select_platform(self, platform_name):
        if isinstance(platform_name, int):
            self._selected_flasher = self._flashers[platform_name]
        elif isinstance(platform_name, str):
            for f in self._flashers:
                if platform_name in f.supported_platforms:
                    self._selected_flasher = f
                    break
        else:
            print("Unknown Platform: {}".format(str(platform_name)))
            return
        self._refresh_platform_files()
        self._refresh_flash_methods()

    def get_platform_names(self):
        flasher_boards = list()
        for j in self._flashers:
            flasher_boards.extend(j.supported_platforms)
        return flasher_boards

    def on_flash_button_clicked(self, b):
        if not self.flashable_files_widget.value:
            return
        if not self._selected_flasher:
            return
        flash_zip_file = Path(self._folder, self.flashable_files_widget.value)
        self._flashing_file = unzip_folder(
            flash_zip_file, self._selected_flasher.flash_file_ext
        )

        self._selected_flasher.flash(
            self._flashing_file, method=self.flash_methods_widget.value
        )

    def _refresh(self, b=None):
        if self._dsk is None:
            return
        self.platform_widget.options = self.get_platform_names()

    def create_widget(self):

        self._flashers = [NordicThingyFlasher(), STDeviceFlasher()]

        self.platform_widget = widgets.Dropdown(description="Platform")
        self.description_widget = widgets.Label(layout=Layout(width="66%"))
        self.flashable_files_widget = widgets.Dropdown(
            description="Binary", layout=Layout(width="85%")
        )
        self.flash_methods_widget = widgets.Dropdown(description="Flash Method")
        self.flash_button = Button(description="Flash")
        self.refresh_button = Button(
            icon="refresh", tooltip="Refresh File List", layout=Layout(width="15%")
        )

        self._refresh()

        self.flash_button.on_click(self.on_flash_button_clicked)
        self.refresh_button.on_click(self._refresh_platform_files)

        return VBox(
            [
                HBox(
                    [
                        widgets.interactive(
                            self.select_platform, platform_name=self.platform_widget
                        ),
                        self.description_widget,
                    ]
                ),
                HBox([self.flashable_files_widget, self.refresh_button]),
                HBox([self.flash_methods_widget, self.flash_button]),
            ]
        )


def unzip_folder(path_to_zip_file, file_ext):
    path_string = str(path_to_zip_file.absolute())
    base_name = os.path.basename(path_string[:-4])
    binary_folder = Path(path_string[:-4])
    binary_path = Path(path_string[:-4], base_name + file_ext)

    zip_ref = zipfile.ZipFile(path_to_zip_file, "r")
    zip_ref.extractall(binary_folder)
    zip_ref.close()

    return binary_path



from sensiml.client import SensiML
from sensiml.widgets import DashBoard


try:
    from IPython.core.display import HTML

    display(HTML("<style>.container { width:90% !important; }</style>"))
except:
    pass

name = "sensiml"


__version__ = "2020.1.3"
__all__ = ["SensiML", "DashBoard"]



import os
import json
from pandas import DataFrame, Series

from sensiml.base import utility
from sensiml.datamanager.sandbox import describe_pipeline


class PipelineSeed(object):
    """Base class for an automation pipeline seed object"""

    _id = 0
    _name = ""
    _description = ""
    _pipeline = ""
    _input_contract = {}

    def __init__(self, connection):
        self._connection = connection

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def description(self):
        return self._description

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def input_contract(self):
        return self._input_contract

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = "seed/{0}/".format(self.id)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._name = response_data["name"]
            self._description = response_data["description"]
            self._id = response_data["id"]
            self._pipeline = response_data["pipeline"]
            self._input_contract = response_data["input_contract"]

    def initialize_from_dict(self, input_dictionary):
        """Populates a single seed object from a dictionary of properties from the server."""
        self._id = input_dictionary["id"]
        self._name = input_dictionary["name"]
        self._description = input_dictionary["description"]
        self._pipeline = input_dictionary["pipeline"]
        self._input_contract = input_dictionary["input_contract"]

    def __dict__(self):
        ret = {
            "Name": self.name,
            "Pipeline": self.pipeline,
            "Description": self.description,
            "Inputs": self.input_contract,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df.set_index("Name").reset_index()

    def describe(self, show_params=True):
        """Prints a formatted description of the pipeline seed"""
        describe_pipeline(self.pipeline, show_params=show_params)



import json
from pandas import DataFrame, Series

from sensiml.datamanager.pipelineseed import PipelineSeed
import sensiml.base.utility as utility


class PipelineSeeds:
    """Base class for a collection of pipeline seeds"""

    def __init__(self, connection):
        self._connection = connection
        self.build_pipeline_seeds()

    def __getitem__(self, index):
        return self.seed_list[index]

    def build_pipeline_seeds(self):
        """Populates the seed_list property from the server."""
        self.seed_list = []
        self.seed_dict = {}

        seeds = self.get_seeds()
        for seed in seeds:
            self.seed_dict[seed.name] = seed
            self.seed_list.append(seed)

    def get_seed_by_name(self, name):
        if len(self.seed_list) == 0:
            self.build_pipeline_seeds()
        return self.seed_dict.get(name, None)

    def _new_seed_from_dict(self, dict):
        """Creates and populates a seed from a set of properties.

            Args:
                dict (dict): contains properties of a Seed

            Returns:
                pipeline seed
        """
        seed = PipelineSeed(self._connection)
        seed.initialize_from_dict(dict)
        return seed

    def get_seeds(self):
        """Gets all seeds as seed objects.

            Returns:
                list of functions
        """
        url = "seed/v2/"
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        pipeline_seeds = []
        for seed in response_data:
            pipeline_seeds.append(self._new_seed_from_dict(seed))

        return pipeline_seeds

    def __call__(self):
        all_seeds = self.get_seeds()
        if len(all_seeds) < 0:
            return DataFrame()
        ret = DataFrame(columns=["Name", "Description"])
        for seed in all_seeds:
            ret = ret.append(seed(), ignore_index=True)
        ret = ret.set_index("Name").reset_index()
        return ret

    def __str__(self):
        output_string = (
            "Seeds are pipeline templates on the server that can be used to generate models \n"
            + "using the dsk.pipeline.auto() command. They eliminate the need to specify a \n"
            + "pipeline and are most effective when chosen appropriately for the application.\n\n"
        )

        for seed in sorted(self.seed_list, key=lambda x: x.id):
            output_string += "Name:          {0}\n".format(seed.name)
            pipeline_format = "{} -> " * len(seed.pipeline)
            output_string += "Pipeline:      {0}\n\n".format(
                pipeline_format.format(*[step["name"] for step in seed.pipeline])[:-3]
            )
            output_string += "{0}\n\n".format(seed.description)

        output_string += "\n"

        return output_string



import os
import IPython
from ipywidgets import widgets
from ipywidgets import (
    Layout,
    Button,
    Box,
    FloatText,
    Textarea,
    Dropdown,
    Label,
    IntSlider,
    Checkbox,
    Text,
    Button,
    SelectMultiple,
)
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget

title_item_layout = Layout(
    display="flex",
    flex_flow="row",
    size=16,
    border="solid 2px",
    justify_content="space-between",
    overflow="visible",
)

label_value_layout = Layout(overflow="visible", min_width="150px")

category_item_layout = Layout(
    display="flex",
    flex_flow="row",
    size=16,
    border="solid 2px",
    justify_content="center",
    background_color="red",
    overflow="visible",
)

form_item_layout = Layout(
    display="flex", flex_flow="row", overflow="visible", justify_content="space-between"
)


pipeline_layout = Layout(
    display="flex",
    flex_flow="column",
    border="solid 2px",
    align_items="stretch",
    width="75%",
    overflow="visible",
)

GENERATE_INDEX = 0
PLATFORM_INDEX = 1
DEBUG_INDEX = 2
TEST_DATA_INDEX = 3
APPLICATION_INDEX = 4
SAMPLE_RATE_INDEX = 5
DOWNLOAD_TYPE = 6
TEST_DATA = 7
PARENT_INDEX = 8
SUBMODEL_INDEX = 9


def clean_name(name):
    return "".join(e if e.isalnum() else "_" for e in name)


class HierarchicalWidget(BaseWidget):
    def __init__(self, dsk=None, level="Project", folder="knowledgepacks"):
        self._dsk = dsk
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)

    def setup(self, level):
        self.kb_description = {"parent": {}, "sub": {"Report": "Report"}}

        self.kb_dict = {"parent": [], "sub": []}

        self.level = level
        self.platforms = self._dsk.platforms()

    def select_platform(self, platform_name):
        self.platform = self.platforms.iloc[platform_name]
        self.platform_name = self.platform["Software Platform"]

        if "curie" in self.platform["Description"].lower():
            self.target_os_widget.options = ["ISPC"]
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ["ble", "led", "serial"]
        elif "nordic" in self.platform["Description"].lower():
            self.target_os_widget.options = ["NordicSDK"]
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ["ble", "led", "serial"]
            self.app_output_options_widget.value = ["ble"]
        elif "simulator" in self.platform["Description"].lower():
            self.target_os_widget.options = ["x86"]
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ["serial"]

    def generate_description(self, b):
        parent_name = None
        parent_uuid = None
        description = {}
        if not self.kb_description["parent"]:
            print("No Parent Model Set.")
        for key in self.kb_description["parent"]:
            description = {
                clean_name(key): {
                    "uuid": self.kb_description["parent"][key],
                    "results": {},
                    "source": "motion",
                }
            }
            parent_name = clean_name(key)
            parent_uuid = self.kb_description["parent"][key]
        for box in self.kb_dict["parent"][0][1:]:
            description[parent_name]["results"].update(
                {
                    format(box.children[0].description.split("-")[0]): clean_name(
                        box.children[0].value
                    )
                }
            )
        for key in self.kb_description["sub"]:
            if key != "Report":
                sub_description = {
                    "uuid": self.kb_description["sub"][key],
                    "parent": parent_name,
                    "segmenter_from": "parent",
                }
                description.update({clean_name(key): sub_description})

        kp_uuid = parent_uuid
        board_name = self._dsk.platforms[self.platform_widget.value].board_name

        kp_platform = self._dsk.platforms[self.platform_widget.value].id
        kp_debug = self.debug_widget.value
        kp_test_data = self.test_data_widget.value
        kp_download_type = self.download_type_widget.value
        sample_rate = self.imu_sample_widget.value
        output_options = self.app_output_options_widget.value

        if kp_platform >= 3 and kp_platform < 99:  # Nordic Thingy and not simulator
            if "ble" not in output_options:
                output_options.insert("ble")  # always output via ble
            kp_application = "thingy_pme"
        else:
            if "led" in output_options:
                kp_application = "LED"
            else:
                kp_application = "Default"

        if board_name == "ARM GCC Generic" or board_name == "x86 GCC Generic":
            kp_application = "testdata_runner"

        if kp_uuid is not None:
            kp = self._dsk.get_knowledgepack(kp_uuid)
        else:
            # print('Invalid KnowledgePack name.')
            return None

        config = {
            "target_platform": kp_platform,
            "test_data": kp_test_data,
            "debug": kp_debug,
            "application": kp_application,
            "sample_rate": sample_rate,
            "output_options": output_options,
            "kb_description": description,
        }

        # print(config)

        if kp_download_type == "Library":
            kp.download_library(config=config, folder="knowledgepacks")
        if kp_download_type == "Binary":
            kp.download_binary(config=config, folder="knowledgepacks")

    def set_parent_model(self, b):
        index = int(b.icon)

        kp = self._dsk.get_knowledgepack(self.kb_items[PARENT_INDEX].children[1].value)
        kp_item = []
        kp_item.append(
            Box(
                [Label(value=kp.name), Button(description="delete")],
                layout=title_item_layout,
            )
        )

        for key, value in kp.class_map.items():
            kp_item.append(
                Box(
                    [
                        Dropdown(
                            options=self.kb_description["sub"],
                            description="{} - {}".format(key, value),
                        )
                    ],
                    layout=form_item_layout,
                )
            )

        self.kb_description["parent"] = {kp.name: kp.uuid}
        self.kb_dict["parent"] = [kp_item]
        self.generate_pipeline()

    def add_submodel(self, b):
        index = int(b.icon)
        kp = self._dsk.get_knowledgepack(
            self.kb_items[SUBMODEL_INDEX].children[1].value
        )
        kp_item = []
        if self.kb_description["sub"].get(kp.name, None) is not None:
            return

        if self.kb_description["parent"].get(kp.name, None) is not None:
            return

        kp_item.append(
            Box(
                [Label(value=kp.name), Button(description="delete")],
                layout=title_item_layout,
            )
        )

        self.kb_description["sub"].update({kp.name: kp.uuid})
        self.kb_dict["sub"].append(kp_item)
        self.generate_pipeline()

    def get_kp_dict(self):
        if self.level.lower() == "project":
            return {
                name: value
                for name, value in self._dsk.project.list_knowledgepacks()[
                    ["Name", "kp_uuid"]
                ].values
                if name
            }
        elif self.level.lower() == "pipeline":
            return {
                name: value
                for name, value in self._dsk.pipeline.list_knowledgepacks()[
                    ["Name", "kp_uuid"]
                ].values
                if name
            }

        return {
            name: value
            for name, value in self._dsk.list_knowledgepacks()[
                ["Name", "kp_uuid"]
            ].values
            if name
        }

    @staticmethod
    def flatten(l):
        return [item for sublist in l for item in sublist]

    def get_model_list(self):
        return (
            [Box([Label(value="Parent Model")], layout=category_item_layout)]
            + self.flatten(self.kb_dict["parent"])
            + [Box([Label(value="Sub Model")], layout=category_item_layout)]
            + self.flatten(self.kb_dict["sub"])
        )

    def get_feature_file_list(self):
        ff = self._dsk.list_featurefiles(silent=True)
        if ff is not None:
            return list(ff["Name"].values)
        else:
            return []

    def get_platform_names(self):
        pf = {}
        for i in range(len(self.platforms)):
            pf[
                "{} {}".format(
                    self.platforms.iloc[i]["Board"],
                    self.platforms.iloc[i]["Platform Version"],
                )
            ] = i

        return pf

    def update(self):

        self.test_data_widget.options = [None] + self.get_feature_file_list()

        self.kb_items[PARENT_INDEX].children[1].options = self.get_kp_dict()
        self.kb_items[SUBMODEL_INDEX].children[1].options = self.get_kp_dict()

        if self.kb_dict["parent"]:
            for child in self.kb_dict["parent"][0][1:]:
                child.children[0].options = [
                    k for k, v in self.kb_description["sub"].items()
                ]

    def create_widget(self):

        self.platform_widget = Dropdown(options=self.get_platform_names())
        self.target_os_widget = Dropdown()
        self.app_output_options_widget = SelectMultiple(
            options=["ble", "led", "serial"]
        )
        self.download_type_widget = Dropdown(options=["Library", "Binary"])
        self.imu_sample_widget = Dropdown(options=[100, 50, 25])
        self.debug_widget = Dropdown(options=[True, False])
        self.test_data_widget = Dropdown(options=[None])

        self.kb_items = [
            Box(
                [
                    Label(value="Model Builder"),
                    Button(icon="0", description="Generate Knowledgepack"),
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Select Hardware Platform", layout=label_value_layout),
                    widgets.interactive(
                        self.select_platform, platform_name=self.platform_widget
                    ),
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Select Target OS", layout=label_value_layout),
                    self.target_os_widget,
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Select Output", layout=label_value_layout),
                    self.app_output_options_widget,
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Download Type", layout=label_value_layout),
                    self.download_type_widget,
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="IMU Sample Rate", layout=label_value_layout),
                    self.imu_sample_widget,
                ],
                layout=form_item_layout,
            ),
            Box(
                [Label(value="Debug", layout=label_value_layout), self.debug_widget],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Test Data", layout=label_value_layout),
                    self.test_data_widget,
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Set Parent Model"),
                    Dropdown(options=[]),
                    Button(icon="1", description="Set"),
                ],
                layout=form_item_layout,
            ),
            Box(
                [
                    Label(value="Add Sub Model"),
                    Dropdown(options=[]),
                    Button(icon="2", description="Add"),
                ],
                layout=form_item_layout,
            ),
        ]

        self.kb_items[GENERATE_INDEX].children[1].on_click(self.generate_description)
        self.kb_items[PARENT_INDEX].children[2].on_click(self.set_parent_model)
        self.kb_items[SUBMODEL_INDEX].children[2].on_click(self.add_submodel)

        model_selector = Box(self.kb_items, layout=pipeline_layout)
        display(model_selector)
        self.generate_pipeline()

    def generate_pipeline(self):
        if self._dsk is None:
            return

        self.update()
        try:
            self._pipeline.close()
        except:
            pass
        self.update()
        self._pipeline = Box(self.get_model_list(), layout=pipeline_layout)
        display(self._pipeline)



from ipywidgets import (
    Layout,
    Button,
    VBox,
    HBox,
    Dropdown,
    Checkbox,
    Label,
    widgets,
    HTML,
    Tab,
)
from sensiml.widgets.base_widget import BaseWidget
import qgrid
import json
from pandas import DataFrame
from sensiml.widgets.renderers import WidgetAttributeRenderer
from sensiml.widgets.feature_generator_widget import FeatureGeneratorWidget
import threading

autosense_metric_column = [
    "Accuracy",
    "Classifer Size(B)",
    "Num. Features",
    "Sensitivity",
    "f1-score",
]


class AutoSenseWidget(BaseWidget):
    def _get_query_list(self):
        q_list = self._dsk.list_queries()
        if q_list is not None:
            return [""] + list(q_list["Name"].values)
        else:
            return [""]

    def _on_button_clicked(self, b):

        self._b_run.disabled = True
        self._b_run.description = "RUNNING..."
        self._b_run.icon = ""
        self._w_project.disabled = True
        self._w_pipeline.disabled = True

        thread = threading.Thread(target=self._async_run_pipeline_optimizer)
        thread.start()

    def _on_stop_button_clicked(self, b):
        if self._dsk and self._dsk.pipeline:
            self._dsk.pipeline.stop_pipeline()
            self.renderer.render("Terminating Pipeline Execution")

    def _async_run_pipeline_optimizer(self):
        try:
            self._run_pipeline_optimizer()
            self._dsk._auto_sense_ran = True
        except:
            self._reset_optimizer_button()

        self._reset_optimizer_button()

    def _reset_optimizer_button(self):

        self._b_run.disabled = False
        self._b_run.description = "Optimize"
        self._b_run.icon = "play"
        self._w_project.disabled = False
        self._w_pipeline.disabled = False

    def _run_pipeline_optimizer(self):
        query_name = self._w_query.value
        segmenter = str(self._w_segment.value)
        seed_choice = str(self._w_seed.value)
        classifiers_sram = self._w_classifiers_sram.value
        iterations = self._w_iterations.value
        population_size = self._w_population_size.value
        mutation_rate = self._w_mutation_rate.value
        recreation_rate = self._w_recreation_rate.value
        survivor_rate = self._w_survivor_rate.value
        reset = self._c_reset.value
        allow_unk = self._c_allow_unk.value
        hierarchical_model = self._c_hierarchical_model.value
        balanced_data = self._c_balanced_data.value
        optimization_metric = self._w_optimization_metric.value.lower()
        demean_segments = self._c_strip_mean.value
        outlier_filter = self._c_outlier_filter.value
        if self._w_combine_labels.value:
            combine_labels = json.loads(self._w_combine_labels.value)
        else:
            combine_labels = None

        if not query_name:
            self.renderer.render("\nERROR: No query name specified!\n")
            return

        if self._dsk.pipeline is None:
            self.renderer.render(
                "\nERROR: Pipeline is not set, use: dsk.pipeline='Pipeline Name'!\n"
            )
            return

        validation_method = {
            "name": self._w_validation_method.value,
            "inputs": {
                "number_of_folds": self._w_number_folds.value,
                "validation_size": self._w_validation_size.value / 100,
                "test_size": 0.0,
                "metadata_name": self._w_validation_metadata.value,
            },
        }

        if optimization_metric == "f1-score":
            optimization_metric = "f1_score"

        params = {
            "iterations": iterations,
            "reset": reset,
            "population_size": population_size,
            "mutation_rate": mutation_rate,
            "recreation_rate": recreation_rate,
            "survior_rate": survivor_rate,
            "allow_unknown": allow_unk,
            "validation_method": validation_method,
            "balanced_data": balanced_data,
            "demean_segments": demean_segments,
            "outlier_filter": outlier_filter,
            "allow_hierarchical_model": hierarchical_model,
            "prediction_target(%)": {optimization_metric: 100},
            "hardware_target": {"classifiers_sram": classifiers_sram},
            "combine_labels": combine_labels,
        }

        add_windowing = False
        delta = None

        if "Windowing" in segmenter:
            delta = self._w_window_size.value
            add_windowing = True
        elif not segmenter:
            self.renderer.render(
                "\n\nERROR: No segmentation algorithm was selected. Select a Segmenter to run this pipeline.\n"
            )
            return

        self._dsk.project.query_optimize()

        self._dsk.pipeline.reset()
        self._dsk.pipeline.set_input_query(query_name)

        if add_windowing:
            self._dsk.pipeline.add_transform(
                "Windowing", params={"window_size": delta, "delta": delta}
            )

        if seed_choice == "Custom Feature Generatorset":
            params[
                "generatorset"
            ] = self.custom_feature_generator.get_fature_generator()

        print("Auto Sense Params", params)
        self._dsk.pipeline.describe()

        self._clear_auto_sense_results()
        self.results, self.summary = self._dsk.pipeline.auto(
            {"seed": seed_choice, "params": params}, renderer=self.renderer
        )

        self.renderer.render("\nAutomation Pipeline Completed.")

        self._set_auto_sense_results()

    def _clear_auto_sense_results(self):
        self._w_results.df = DataFrame(
            [[0, 0, 0, 0, 0] for _ in range(5)], columns=autosense_metric_column
        )

    def _set_auto_sense_results(self):

        self._clear_auto_sense_results()

        if (
            self.summary is not None
            and isinstance(self.summary, dict)
            and isinstance(self.summary.get("fitness_summary", None), DataFrame)
        ):
            if "f1_score" not in self.summary["fitness_summary"].columns:
                self.summary["fitness_summary"]["f1_score"] = 0

            df_results = (
                self.summary["fitness_summary"][
                    [
                        "accuracy",
                        "classifiers_sram",
                        "features",
                        "sensitivity",
                        "f1_score",
                    ]
                ]
                .astype(int)
                .head()
            )
            df_results = df_results.rename(
                index=str, columns={"classifiers_sram": "Classifier Size (B)"}
            )
            df_results = df_results.rename(index=str, columns={"accuracy": "Accuracy"})
            df_results = df_results.rename(
                index=str, columns={"sensitivity": "Sensitivity"}
            )
            df_results = df_results.rename(index=str, columns={"f1_score": "f1-score"})
            df_results = df_results.rename(
                index=str, columns={"features": "Num. Features"}
            )

            self._w_results.df = df_results
        else:
            if isinstance(self.results, dict):
                self.renderer.render(
                    self.results.get("message", "").replace("\n", "<br>")
                )

    def _terminate_run(self, b):
        self._dsk.pipeline.stop_pipeline()

    def _select_seed(self, Name):
        if self._dsk and Name:
            self._w_seed_desc.value = self._dsk.seeds.get_seed_by_name(Name).description
            if Name == "Custom Feature Generatorset":
                self.custom_feature_generator.show()
            else:
                self.custom_feature_generator.hide()

    def _on_value_change(self, change):
        if self._dsk is None:
            return
        if self._dsk.pipeline and change["new"]:
            self._dsk.pipeline.reset()
            self._dsk.pipeline.set_input_query(change["new"])
            self._w_validation_metadata.options = [
                x
                for x in self._dsk.pipeline.group_columns
                if x not in [self._dsk.pipeline.label_column, "SegmentID"]
            ]
            query = self._dsk.project.queries.get_or_create_query(change["new"])
            if self._dsk.project.get_segmenters().loc[query.segmenter]["custom"]:
                self._w_segment.options = ["", "Windowing", "Manual Segments"]
            else:
                self._w_segment.options = ["", "Windowing", "Query Segmenter"]

        else:
            self.renderer.render("No Pipeline Selected.")

    def _on_validation_method_change(self, change):
        def set_states_visible(widgets):
            for widget in widgets:
                widget.layout.visibility = "visible"
                widget.layout.height = "auto"

        def set_states_hidden(widgets):
            for widget in widgets:
                widget.layout.visibility = "hidden"
                # widget.layout.height = '1px'

        if change["new"] in ["Stratified Shuffle Split"]:
            set_states_visible(
                [self._w_validation_size, self._info_validation_method_size]
            )
        else:
            set_states_hidden(
                [self._w_validation_size, self._info_validation_method_size]
            )

        if change["new"] in ["Stratified Metadata k-fold", "Metadata k-fold"]:
            set_states_visible(
                [self._info_validation_metadata, self._w_validation_metadata]
            )
        else:
            set_states_hidden(
                [self._info_validation_metadata, self._w_validation_metadata]
            )

        if change["new"] in [
            "Stratified K-Fold Cross-Validation",
            "Stratified Shuffle Split",
            "Metadata k-fold",
            "Stratified Metadata k-fold",
        ]:
            set_states_visible(
                [self._w_number_folds, self._info_validation_method_folds]
            )
        else:
            set_states_hidden(
                [self._w_number_folds, self._info_validation_method_folds]
            )

    def _on_segmenter_change(self, change):
        if change["new"] == "Windowing":
            self._w_window_size.disabled = False
            self._w_window_size.layout.visibility = "visible"
            self._info_window_size.layout.visibility = "visible"
        else:
            self._w_window_size.disabled = True
            self._w_window_size.layout.visibility = "hidden"
            self._info_window_size.layout.visibility = "hidden"

    def _refresh(self, b=None):
        if self._dsk:
            self._w_query.options = self._get_query_list()
            self._w_query.values = self._w_query.options[0]
            self._w_seed.options = sorted(self._dsk.seeds.seed_dict.keys())

            if self._dsk.pipeline:
                self._w_pipeline_label.value = (
                    "Pipeline Name: <b>" + self._dsk.pipeline.name + "</b>"
                )
                self.results, self.summary = self._dsk.pipeline.get_results(
                    renderer=self.renderer
                )
                self._set_auto_sense_results()
                self.custom_feature_generator._dsk = self._dsk
                self.custom_feature_generator._refresh()

    def _clear(self):
        self._w_query.options = [""]
        self._w_query.value = ""
        self._w_validation_metadata.options = []

    def create_widget(self):

        info_layout = Layout(width="35px", visibility="visible")

        description_style = {"description_width": "112px"}

        self._w_pipeline_label = HTML(value="Pipeline Name:")
        self._w_place_holder = Label(value="")

        self._w_query = Dropdown(
            description="Query", options=[], style=description_style
        )

        self._w_segment = widgets.Dropdown(
            description="Segmenter",
            options=["", "Windowing"],
            style=description_style,
        )

        self._w_window_size = widgets.BoundedIntText(
            value=1024,
            min=25,
            max=8192,
            step=100,
            description="Window Size:",
            disabled=True,
            style=description_style,
        )

        self._info_segmenter = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="""
                    Windowing: Applies a sliding windowing to the segments you have labeled in the Data Capture Lab.
                     Every segment that is created this way is assigned the class associated with the segment it was
                     created from.
                    
                    Query Segmenter: If you used a built in segmentation algorithm, this will use that as well
                     as any preprocess steps as the first part of the pipeline. Each segment will be assigned the 
                     class label it was given in the Data Capture lab.

                    Manual Segments: In this case, the raw segments are fed into the algorithm without any additional
                    modification. If you use this method you will not be able to compile firmware directly but will 
                    have to build a library and create your own custom segmentation algorithm. 
                    """,
            layout=info_layout,
        )

        tooltip_text = """
                    Accuracy : Ability to differentiate events correctly. 
                    Accuracy as an good evaluation metrics if the class labels are uniformly distributed.
                    For more detail, see https://en.wikipedia.org/wiki/Accuracy_and_precision
                    
                    f1 score : Helps to measure sensitivity and ability to predict individual events correctly out of all the positive prediction at the same time. 
                    f1 Score might be a better statistical metric to use if there is an uneven class distribution. It seeks a balance between sensitivity and positive prediction.
                    For more detail, see https://en.wikipedia.org/wiki/F1_score            

                    Sensitivity : Ability to predict individual events correctly out of all the positive events.
                    With imbalanced classes, accuracy can give you false assumptions regarding the classifier’s performance.
                    Therefore, using sensitivity is suggested for data sets with imbalanced classes.  
                    For more detail, see https://en.wikipedia.org/wiki/Sensitivity_and_specificity
            """.replace(
            "  ", ""
        )

        self._info_optimization_metric = widgets.Button(
            icon="question", disabled=True, tooltip=tooltip_text, layout=info_layout
        )

        self._info_classifiers_sram = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Defines the maximum SRAM usage of the classifier parameters for the model.",
            layout=info_layout,
        )

        self._info_validation_method = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Select the validation method that will be used to calculate the performance metrics such as accuracy and sensitivity of the model.",
            layout=info_layout,
        )

        self._info_validation_method_size = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Select the percentage of data to hold out in the validation folds.",
            layout=Layout(width="35px"),
        )

        self._info_validation_method_folds = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Select the number of folds for validation .",
            layout=Layout(width="35px"),
        )

        self._info_validation_metadata = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Select the metadata to group by.",
            layout=Layout(width="35px"),
        )

        self._info_balanced_data = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Balance the number of examples from each class by undersampling the majority classes. This will reduce number of examples of each class to match the class with the smallest number of examples in your data set.",
            layout=info_layout,
        )

        self._info_allow_unk = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Uses classifiers which will return unknown when the class falls outside of the known predictive space.",
            layout=info_layout,
        )

        self._info_hierarchical_model = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Optimize the prediction results by creating Hierarchical Binary Models.",
            layout=info_layout,
        )

        self._info_population = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Defines how large the initial population is. A higher population means a larger initial search space is.\nA higher population typically produces better results but takes more time.",
            layout=info_layout,
        )

        self._info_iterations = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Defines the number of iterations the model will go through.\n At each iteration a new population of models is created by mutating the previous iterations population.\nA higher number of iteration produces better results but takes more time.",
            layout=info_layout,
        )

        self._info_recreation_rate = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Rate at which new populations are created during each iteration",
            layout=info_layout,
        )

        self._info_survivor_rate = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Rate at which populations survive through each iteration",
            layout=info_layout,
        )

        self._info_mutation_rate = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Rate at which populations mutate through each iteration",
            layout=info_layout,
        )

        self._info_strip_mean = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Remove the mean from  each data source before before extracting features. This helps avoid overfitting models to noise and drift.",
            layout=info_layout,
        )

        self._info_outlier_filter = widgets.Button(
            icon="question",
            disabled=True,
            tooltip="Filter outliers for each class using Isolation forest filtering. Filters 5pct. of outliers",
            layout=info_layout,
        )

        tooltip_window_size = """
            The number of samples to buffer for each event before extracting features and
            performing classification. The window size can have a large effect on model 
            accuracy as well as the size of the resulting knowledge pack. A larger window 
            size will allow your model to capture longer trends at the expense of shorter
            range information. A larger window will also need to store more data for 
            processing which will increase the SRAM as well as the latency.""".replace(
            "  ", ""
        )

        self._info_window_size = widgets.Button(
            icon="question",
            disabled=True,
            tooltip=tooltip_window_size,
            layout=Layout(width="35px"),
        )

        self._w_seed = widgets.Dropdown(
            description="Feature Family", options=[], style=description_style
        )

        self._w_seed_desc = widgets.HTML(
            description="Description", disable=True, style=description_style
        )

        self._w_intereact = widgets.interactive(self._select_seed, Name=self._w_seed)

        self._w_classifiers_sram = widgets.BoundedIntText(
            description="Classifier Size (B)",
            value=32000,
            min=0,
            max=512000,
            style=description_style,
        )

        self._w_iterations = widgets.IntSlider(
            description="Iterations",
            value=4,
            min=1,
            max=15,
            step=1,
            style=description_style,
        )

        self._w_population_size = widgets.IntSlider(
            description="Initial Population",
            value=100,
            min=25,
            max=200,
            step=1,
            style=description_style,
        )

        self._c_reset = Checkbox(description="Reset", value=True)

        self._c_allow_unk = Checkbox(description="Allow Unknown", value=False)

        self._c_hierarchical_model = Checkbox(
            description="Hierarchical Model Optimization", value=False
        )

        self._c_balanced_data = Checkbox(description="Balance Data", value=False)

        self._c_strip_mean = Checkbox(description="Strip Mean", value=False)

        self._c_outlier_filter = Checkbox(description="Filter Outliers", value=False)

        self._w_combine_labels = widgets.Text(description="Combine Labels", value="")

        self._w_validation_method = widgets.Dropdown(
            description="Validation",
            options=[
                "Stratified K-Fold Cross-Validation",
                "Stratified Shuffle Split",
                "Metadata k-fold",
                "Stratified Metadata k-fold",
                "Recall",
            ],
            style=description_style,
        )

        self._w_validation_size = widgets.IntSlider(
            description="Validation(%)",
            value=20,
            min=0,
            max=100,
            step=10,
            style=description_style,
        )

        self._w_validation_size = widgets.IntSlider(
            description="Validation(%)",
            value=20,
            min=0,
            max=100,
            step=10,
            style=description_style,
        )

        self._w_validation_metadata = widgets.Dropdown(
            description="metadata", options=[], style=description_style
        )

        self._w_number_folds = widgets.IntSlider(
            description="# Folds",
            value=5,
            min=2,
            max=10,
            step=1,
            style=description_style,
        )

        self._w_mutation_rate = widgets.FloatSlider(
            description="Mutation Rate",
            value=0.1,
            min=0,
            max=1,
            step=0.05,
            style=description_style,
        )

        self._w_recreation_rate = widgets.FloatSlider(
            description="Recreation Rate",
            value=0.2,
            min=0,
            max=1,
            step=0.05,
            style=description_style,
        )

        self._w_survivor_rate = widgets.FloatSlider(
            description="Survivor Rate",
            value=0.5,
            min=0,
            max=1,
            step=0.05,
            style=description_style,
        )

        self._w_optimization_metric = widgets.Dropdown(
            description="Optimization Metric",
            options=["Accuracy", "f1-score", "Sensitivity"],
            style=description_style,
        )

        self._w_results = qgrid.show_grid(
            DataFrame(
                [[0, 0, 0, 0, 0] for _ in range(5)], columns=autosense_metric_column
            ),
            grid_options={
                "rowHeight": 48,
                "maxVisibleRows": 6,
                "minVisibleRows": 6,
                "editable": False,
                "defaultColumnWidth": 50,
                "forceFitColumns": True,
                "sortable": True,
                "filterable": False,
            },
        )

        self._w_results.layout.width = "95%"
        self._w_results.layout.align_self = "flex-end"

        # , layout=Layout( width='98%', align_self='flex-end'))
        self._b_run = Button(
            icon="play",
            description="Optimize",
            layout=Layout(align_self="center", width="250px"),
        )
        self._b_stop = Button(
            icon="stop",
            description="Stop",
            layout=Layout(align_self="center", width="250px"),
        )
        self._b_refresh = Button(
            icon="refresh", layout=Layout(width="35px", align_self="flex-end")
        )
        self._b_terminate = Button(
            icon="stop",
            description="Terminate",
            layout=Layout(width="25%", align_self="flex-end"),
        )
        self._b_iterate = Button(
            icon="redo",
            description="Iterate",
            layout=Layout(width="98%", align_self="flex-end"),
        )

        self._widget_render_space = HTML()

        self.renderer = WidgetAttributeRenderer(self._widget_render_space, "value")

        self._w_query.observe(self._on_value_change, names="value")
        self._b_run.on_click(self._on_button_clicked)
        self._b_stop.on_click(self._on_stop_button_clicked)
        self._b_refresh.on_click(self._refresh)
        self._b_run.style.button_color = "#90EE90"
        self._b_terminate.on_click(self._terminate_run)
        self._w_validation_method.observe(
            self._on_validation_method_change, names="value"
        )
        self._w_segment.observe(self._on_segmenter_change, names="value")

        self._w_window_size.layout.visibility = "hidden"
        self._info_window_size.layout.visibility = "hidden"
        self._w_validation_size.layout.visibility = "hidden"
        self._info_validation_method_size.layout.visibility = "hidden"
        self._info_validation_metadata.layout.visibility = "hidden"
        self._w_validation_metadata.layout.visibility = "hidden"

        self._refresh()
        self.custom_feature_generator = FeatureGeneratorWidget(None)
        self.custom_feature_generator.create_widget()
        self.custom_feature_generator.hide()

        autosense_tabs = Tab(
            [
                HBox(
                    [
                        VBox(
                            [
                                self._w_pipeline_label,
                                widgets.HTML("<b>Settings</b>"),
                                HBox([self._w_query, self._b_refresh]),
                                HBox([self._w_segment, self._info_segmenter]),
                                HBox([self._w_window_size, self._info_window_size]),
                                widgets.HTML("<b>Target Constraints</b>"),
                                HBox(
                                    [
                                        self._w_optimization_metric,
                                        self._info_optimization_metric,
                                    ]
                                ),
                                HBox(
                                    [
                                        self._w_classifiers_sram,
                                        self._info_classifiers_sram,
                                    ]
                                ),
                                self._w_place_holder,
                                HBox([self._b_run, self._b_stop]),
                            ],
                            layout=Layout(width="800px"),
                        ),
                        VBox(
                            [
                                self._w_place_holder,
                                self._w_results,
                                self._widget_render_space,
                            ]
                        ),
                    ],
                    layout=Layout(
                        width="100%",
                        display="flex",
                        flex_flow="row",
                        justify_content="space-between",
                    ),
                ),
                VBox(
                    [
                        widgets.HTML("<b>Pipeline Settings</b>"),
                        HBox(
                            [self._c_strip_mean, self._info_strip_mean],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._c_balanced_data, self._info_balanced_data],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._c_hierarchical_model, self._info_hierarchical_model],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._c_outlier_filter, self._info_outlier_filter],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._c_allow_unk, self._info_allow_unk],
                            layout=Layout(align_self="flex-start"),
                        ),
                        # HBox([self._w_combine_labels],
                        #     layout=Layout(align_self='flex-start')),
                        widgets.HTML("<b>Feature Extractors</b>"),
                        self._w_intereact,
                        self._w_seed_desc,
                        self.custom_feature_generator._w_grid,
                        widgets.HTML("<b>Validation Settings</b>"),
                        HBox(
                            [self._w_validation_method, self._info_validation_method],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._w_number_folds, self._info_validation_method_folds],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [
                                self._w_validation_size,
                                self._info_validation_method_size,
                            ],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [
                                self._w_validation_metadata,
                                self._info_validation_metadata,
                            ],
                            layout=Layout(align_self="flex-start"),
                        ),
                    ]
                ),
                VBox(
                    [
                        widgets.HTML("<b>Genetic Search Setttings</b>"),
                        HBox(
                            [self._w_population_size, self._info_population],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._w_iterations, self._info_iterations],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._w_recreation_rate, self._info_recreation_rate],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._w_mutation_rate, self._info_mutation_rate],
                            layout=Layout(align_self="flex-start"),
                        ),
                        HBox(
                            [self._w_survivor_rate, self._info_survivor_rate],
                            layout=Layout(align_self="flex-start"),
                        ),
                    ]
                ),
            ]
        )

        autosense_tabs.set_title(0, "Main Settings")
        autosense_tabs.set_title(1, "Pipeline Settings")
        autosense_tabs.set_title(2, "Advanced Settings")

        self._widget = VBox([autosense_tabs])

        return self._widget



from sensiml.datamanager.pipeline import PipelineStep


class CaptureFileCall(PipelineStep):
    """The base class for a featurefile call"""

    def __init__(self, name):
        super(CaptureFileCall, self).__init__(name=name, step_type="CaptureFileCall")
        self.name = name
        self._data_columns = None
        self._group_columns = ["Subject"]

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, list):
            name = [name]
        self._name = name

    @property
    def data_columns(self):
        return self._data_columns

    @data_columns.setter
    def data_columns(self, value):
        self._data_columns = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    @property
    def group_columns(self):
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        print("Capture files default to Subject.")

    def _to_dict(self):
        capturefile_dict = {}
        capturefile_dict["name"] = self.name
        capturefile_dict["type"] = "capturefile"
        capturefile_dict["data_columns"] = getattr(self, "_data_columns")
        capturefile_dict["group_columns"] = getattr(self, "_group_columns")
        capturefile_dict["outputs"] = getattr(self, "_outputs")
        return capturefile_dict



import json
from sensiml.datamanager.featurefile import FeatureFile
import sensiml.base.utility as utility


class FeatureFileExistsError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class FeatureFiles:
    """Base class for a collection of featurefiles."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project

    def create_featurefile(self, filename, path):
        """Creates a featurefile object from the filename and path.

        Args:
            filename (str): desired name of the featurefile on the server, must have a .csv or .arff extension
            path (str): full local path to the file, including the file's local name and extension

        Returns:
            featurefile object

        Raises:
            FeatureFileExistsError, if the featurefile already exists on the server
        """
        if self.get_featurefile_by_name(filename) is not None:
            raise FeatureFileExistsError(
                "featurefile {0} already exists.".format(filename)
            )
        else:
            featurefile = self.new_featurefile()
            featurefile.name = filename
            featurefile.path = path
            featurefile.insert()
            return featurefile

    def build_featurefile_list(self):
        """Populates the function_list property from the server."""
        featurefile_list = {}

        featurefile_response = self.get_featurefiles()
        for featurefile in featurefile_response:
            featurefile_list[featurefile.name] = featurefile

        return featurefile_list

    def get_featurefile_by_name(self, filename):
        """Gets a featurefile from the server.

        Args:
            filename: name of the featurefile as stored on the server

        Returns:
            featurefile object or None if it does not exist
        """
        featurefile_list = self.build_featurefile_list()
        return featurefile_list.get(filename, None)

    def new_featurefile(self):
        """Initializes a new featurefile object, but does not insert it."""
        featurefile = FeatureFile(self._connection, self._project)
        return featurefile

    def _new_featurefile_from_dict(self, dict):
        """Creates a featurefile object from a dictionary of properties.

        Args:
            dict (dict): contains featurefile's 'name' and 'uuid' properties

        Returns:
            featurefile object
        """
        featurefile = FeatureFile(self._connection, self._project)
        featurefile.initialize_from_dict(dict)
        return featurefile

    def get_featurefiles(self):
        """Gets a list of all featurefiles in the project.

        Returns:
            list (featurefiles)
        """
        err = False
        url = "project/{0}/featurefile/".format(self._project.uuid)
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)
        # Populate the retrieved featurefiles
        featurefiles = []
        if err is False:
            try:
                for featurefile_params in response_data:
                    featurefiles.append(
                        self._new_featurefile_from_dict(featurefile_params)
                    )
            except:
                pass

        return featurefiles



import json
from sensiml.datamanager.metadata import Metadata
import sensiml.base.utility as utility


class MetadataExistsError(Exception):
    """Base class for the metadata exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class MetadataSet:
    """Base class for a collection of metadata"""

    def __init__(self, connection, project, capture):
        self._connection = connection
        self._project = project
        self._capture = capture

    def __getitem__(self, name):
        metadata = self.get_metadata_by_name(name)
        if metadata is not None:
            return metadata.get_metadata_value()
        else:
            return None

    def __setitem__(self, name, value):
        metadata = self.get_metadata_by_name(name)
        if metadata is not None:
            metadata.set_metadata_value(value)
            metadata.update()
        else:
            self.create_metadata(name, value)

    def create_metadata(self, name, capture):
        """Creates a metadata object using its name and capture properties.

            Args:
                name (str): name of the metadata object
                capture (capture): parent capture of the metadata
        """
        # TODO: This looks odd with capture having a _metadata attribute; make sure this is being unit tested
        if self.get_metadata_by_name(name) is not None:
            raise MetadataExistsError("Metadata {0} already exists.".format(name))
        else:
            self._capture = capture
            self._capture.await_ready()
            metadata = self.new_metadata()
            metadata.name = name
            metadata.new_value = capture._metadata.new_value
            metadata.insert()
            return  # metadata

    def get_metadata_by_name(self, name):
        """Gets one metadata object by the name property.

            Args:
                name (str)

            Returns:
                metadata object or None

            Note:
                If more than one metadata of the same name exist for the capture, this will only return one of them
                at random.
        """
        metadata_list = self.get_metadataset()
        for metadata in metadata_list:
            if metadata.name == name:
                return metadata
        return None

    def new_metadata(self):
        """Initializes a new metadata object locally but does not insert it.

            Returns:
                metadata object
        """
        metadata = Metadata(self._connection, self._project, self._capture)
        return metadata

    def _new_metadata_from_dict(self, dict):
        """Creates a new metadata from data in the dictionary.

            Args:
                dict (dict): contains metadata properties uuid, name, type, value,

            Returns:
                metadata object

        """
        metadata = Metadata(self._connection, self._project, self._capture)
        metadata.initialize_from_dict(dict)
        return metadata

    def get_metadataset(self):
        """Gets all of the capture's metadata objects in a list.

            Returns:
                list[metadata]
        """
        # Query the server and get the json
        url = "v2/project/{0}/capture/{1}/metadata-relationship/".format(
            self._project.uuid, self._capture.uuid
        )
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        # Populate each metadata from the server
        metadataset = []
        if err is False:
            try:
                for metadata_params in response_data:
                    metadataset.append(self._new_metadata_from_dict(metadata_params))
            except:
                pass
        return metadataset



import json
from numpy import float64, int64
import sensiml.base.utility as utility
import logging
import time


logger = logging.getLogger(__name__)


class LabelValue(object):
    Int = "integer"
    Float = "float"
    String = "string"


class Segment(object):
    """Base class for a label object."""

    def __init__(self, connection, project, capture, segmenter=None, label=None, label_value=None):
        """Initialize a metadata object.

            Args:
                connection
                project
                capture
        """
        self._uuid = ""
        self._sample_start = 0
        self._sample_end = 0
        self._connection = connection
        self._project = project
        self._capture = capture
        self._segmenter = segmenter
        self._label = label
        self._label_value = label_value

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value



    @property
    def sample_start(self):
        """The index of the first sample of the label"""
        return self._sample_start

    @sample_start.setter
    def sample_start(self, value):
        self._sample_start = value

    @property
    def sample_end(self):
        """The index of the last sample of the label"""
        return self._sample_end

    @sample_end.setter
    def sample_end(self, value):
        self._sample_end = value

    @property
    def segmenter(self):
        """The index of the last sample of the label"""
        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        self._segmenter = value

    @property
    def label(self):
        if isinstance(self._label, str):
            return self._label
        else:
            return self._label.uuid

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def label_value(self):
        if isinstance(self._label_value, str):
            return self._label_value
        else:
            return self._label_value.uuid

    @label_value.setter
    def label_value(self, value):
        self._label_value = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        self._capture.await_ready()
        url = "project/{0}/capture/{1}/label-relationship/".format(
            self._project.uuid, self._capture.uuid
        )
        label_info = {
            "capture_sample_sequence_start": self.sample_start,
            "capture_sample_sequence_end": self.sample_end,
            "label": self.label,
            "label_value": self.label_value,
            "segmenter": self.segmenter,
        }
        response = self._connection.request("post", url, label_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

    def update(self):
        """Calls the REST API and updates the object on the server."""
        self._capture.await_ready()
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        label_info = {
            "capture_sample_sequence_start": self.sample_start,
            "capture_sample_sequence_end": self.sample_end,
            "label": self.label,
            "label_value": self.label_value,
            "segmenter": self.segmenter,
        }
        response = self._connection.request("put", url, label_info)
        response_data, err = utility.check_server_response(response)

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

            Args:
                data (dict): contains the uuid, type, value, capture_sample_sequence_start, and
                capture_sample_sequence_end properties
        """
        self.uuid = data["uuid"]
        self.sample_start = data["capture_sample_sequence_start"]
        self.sample_end = data["capture_sample_sequence_end"]
        self.segmenter = data.get("segmenter", None)
        self.label = data['label']
        self.label_value = data['label_value']
        

class SegmentSet:
    """Base class for a collection of segments"""

    def __init__(self, connection, project, capture, initialize_set=True):
        self._connection = connection
        self._project = project
        self._capture = capture
        self._set = None

        if initialize_set:
            self.refresh()
        

    @property
    def segments(self):
        if self._set is None:
            self._set = self.get_set()

        return self._set

    def refresh(self):
        self._set = self.get_set()

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

            Args:
                dict (dict): contains label properties uuid, name, type, value, capture_sample_sequence_start, and
                capture_sample_sequence_end

            Returns:
                label object

        """
        segment = Segment(self._connection, self._project, self._capture)
        segment.initialize_from_dict(data)
        return segment

    def get_set(self):
        """Gets all of the capture's label objects in a list.

            Returns:
                list[label]
        """
        # Query the server and get the json
        url = "project/{0}/capture/{1}/label-relationship/".format(
            self._project.uuid, self._capture.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err:
            print(err)
            return None

        # Populate each label from the server
        segmentset = []

        for obj in response_data:
            segmentset.append(self._new_obj_from_dict(obj))

        return segmentset



import json
from sensiml.datamanager.capture import Capture
import sensiml.base.utility as utility


class CaptureExistsError(Exception):
    """Base class for an Capture exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Captures:
    """Base class for a collection of Captures."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project
        self._reserved_metadata_names = ["capture_uuid", "segment_uuid"]

    def create_capture(self, filename, filepath, asynchronous=True, capture_info=None):
        """Creates an capture object from the given filename and filepath.

            Args:
                filename (str): desired name of the file on the server
                filepath (str): local path to the file to be uploaded
                asynchronous (bool): Whether to process asynchronously

            Returns:
                capture object

            Raises:
                CaptureExistsError, if the Capture already exists on the server
        """
        capture_dict = {
            "connection": self._connection,
            "project": self._project,
            "filename": filename,
        }

        if self.get_capture_by_filename(filename) is not None:
            raise CaptureExistsError("capture {0} already exists.".format(filename))
        else:
            capture = Capture.initialize_from_dict(capture_dict)
            capture.path = filepath
            capture.capture_info = (
                capture_info if isinstance(capture_info, dict) else {}
            )
            capture.insert(asynchronous=asynchronous)

        return capture

    def build_capture_list(self):
        """Populates the function_list property from the server."""
        capture_list = {}

        capture_response = self.get_captures()
        for capture in capture_response:
            capture_list[capture.filename] = capture

        return capture_list

    def get_capture_by_filename(self, filename):
        """Gets an capture object from the server using its filename property.

            Args:
                filename (str): the capture's name

            Returns:
                capture object or None if it does not exist
        """
        capture_list = self.get_captures()
        for capture in capture_list:
            if capture.filename == filename:
                return capture
        return None

    def get_capture_by_uuid(self, uuid):
        """Gets an capture object from the server using its filename property.

            Args:
                filename (str): the capture's name

            Returns:
                capture object or None if it does not exist
        """
        """Calls the REST API to update the capture."""
        url = "project/{0}/capture/{1}/".format(self._project.uuid, uuid)

        response = self._connection.request("get", url)       
        response_data, err = utility.check_server_response(response)

        return self._new_capture_from_dict(response_data)


    def __getitem__(self, key):
        if type(key) == str:
            return self.get_capture_by_filename(key)
        else:
            return self.get_captures()[key]

    def get_statistics(self):
        """Gets all capture statistics for the project.

            Returns:
                DataFrame of capture statistics
        """
        url = "project/{0}/statistics/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            data, error_report = utility.make_statistics_table(response_data)
            return data, error_report
        else:
            return None

    def get_metadata_names_and_values(self):
        """Gets all the metadata names and possible values for a project.

            Returns:
                list(dict) containing metadata names and values
        """
        url = "project/{0}/metadata/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [
                {
                    "name": item["name"],
                    "values": [i["value"] for i in item["label_values"]],
                }
                for item in response_data
            ]
            return metadata_values

    def get_metadata_names(self):
        """Gets all the metadata names within a project.

            Returns:
                list of metadata names
        """
        url = "project/{0}/metadata/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [
                item["name"] for item in response_data
            ] + self._reserved_metadata_names

    def get_captures_by_metadata(self, key, value):
        """Gets captures by existing metadata key-values.

            Args:
                key (str): the name of the metadata item
                value (str, int, or float): the value to search for

            Returns:
                list of captures that have the desired metadata key-value pair
        """
        url = "project/{0}/capture/metadata/[{1}]=[{2}]/".format(
            self._project.uuid, key, value
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            captures = []
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))
            return captures

    def get_label_names_and_values(self):
        """Gets all the label names and possible values for a project.

            Returns:
                list(dict) containing metadata names and values
        """
        url = "project/{0}/label/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [
                {
                    "name": item["name"],
                    "values": [i["value"] for i in item["label_values"]],
                }
                for item in response_data
            ]
            return metadata_values

    def get_label_names(self):
        """Gets all the label names within a project.

            Returns:
                list of label names
        """
        url = "project/{0}/label/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [item["name"] for item in response_data]

    def _new_capture_from_dict(self, capture_dict):
        """Creates a new capture using the dictionary.

            Args:
                capture_dict (dict): dictionary of capture properties

            Returns:
                capture object
        """
        capture_dict.update({"connection": self._connection, "project": self._project})
        return Capture.initialize_from_dict(capture_dict)

    def get_captures(self):
        """Gets all captures from the server.

            Returns:
                list of captures for the project
        """
        # Query the server and get the json
        url = "project/{0}/capture/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        captures = []
        if err is False:
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))

        return captures



import sys
import os
from setuptools import setup, find_packages

__version__ = "2020.1.3"

# 'setup.py publish' shortcut.
if sys.argv[-1] == "publish":
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    sys.exit()

# 'setup.py test' shortcut.
# !pip install --index-url https://test.pypi.org/simple/ sensiml -U
if sys.argv[-1] == "test":
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload --repository-url https://test.pypi.org/legacy/ dist/*")
    sys.exit()

setup(
    name="SensiML",
    description="SensiML Analytic Suite Python client",
    version=__version__,
    author="SensiML",
    author_email="support@sensiml.com",
    license="Proprietary",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=["*test*"]),
    package_data={
        "sensiml.datasets": ["*.csv"],
        "sensiml.widgets": ["*.pem"],
        "sensiml.image": ["*.png"],
    },
    include_package_data=True,
    long_description=open("README.md").read(),
    install_requires=[
        "cookiejar==0.0.2",
        "requests>=2.14.2",
        "requests-oauthlib>=0.7.0",
        "appdirs==1.4.3",
        "semantic_version>=2.6.0",
        "jupyter>=1.0.0",
        "numpy>=1.13",
        "pandas>=0.20.3,<1.0.0",
        "matplotlib>=2.0.0",
        "nrfutil>=3.3.2,<=5.0.0",
        "qgrid>=1.0.2",
        'prompt-toolkit>=2.0.5; python_version>="3"',
        'prompt-toolkit<=1.0.4; python_version<"3"',
        'jupyter-console>=6.0.0; python_version>="3"',
        'notebook==5.7.5; python_version>="3"',
        'jupyter-console<=5.2.0; python_version<"3"',
        'ipython>=7.0.1; python_version>="3"',
        'ipython<=5.8.0; python_version<"3"',
        'ipywidgets>=7.5.1; python_version>="3"',
        "pywin32==225 ; sys_platform == 'win32'",
        "bqplot",
        "seaborn",
        "wurlitzer",
        "jupyter-contrib-nbextensions",
        "pyserial",
    ],
)



import uuid
import itertools
import copy
import json
from tempfile import NamedTemporaryFile
from os import remove
import time
import logging
from pandas import DataFrame, Series
from requests.exceptions import HTTPError
from six import string_types
import numpy as np

from sensiml.base.exceptions import DuplicateValueError
import sensiml.base.utility as utility
from sensiml.base.snippets import generate_pipeline
from sensiml.base.exceptions import *

try:
    from sensiml.visualize import Visualize
except:
    print("skipping visualization library")
    pass


logger = logging.getLogger(__name__)


class InvalidParameterException(Exception):
    pass


class InvalidModelException(Exception):
    pass


class PlatformNotSpecifiedException(Exception):
    pass


class Pipeline(object):
    """Base class of a pipeline object"""

    def __init__(self, kb, name=None):
        self._kb = kb
        self._project = kb.project
        self._results = None
        self._stats = None
        self._dataset = None
        self._async = None
        self._save = False
        self._tvo_call = None
        self._generator_index = 0
        self._selector_index = 0
        self._classifier_call = None
        self._training_algorithm_call = None
        self._validation_call = None
        self._group_columns = None
        self._label_column = None
        self._data_columns = None

        if name:
            self._id = "{}".format(name)
        else:
            self._id = "{}_sandbox".format(time.time())

        self._sandbox = self._project.sandboxes.get_or_create_sandbox(self._id)

        logger.debug("kb_dsk_pipeline_instance:" + self._id)

    @property
    def name(self):
        return self._id

    @property
    def data_columns(self):
        if self._data_columns:
            return sorted(list(self._data_columns))
        return self._data_columns

    @data_columns.setter
    def data_columns(self, value):
        if isinstance(value, list):
            if len(set(value)) is not len(value):
                raise DuplicateValueError()
            self._data_columns = set(map(lambda x: str(x), value))
        else:
            print("Error: data_columns must be a list.")
            return

        upper_case_data_columns = set(map(lambda x: x.upper(), self.data_columns))

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        if isinstance(value, str) or isinstance(value, string_types):
            self._label_column = str(value)
        else:
            print("Error: label_column must be a string.")

    @property
    def group_columns(self):
        if self._group_columns:
            return sorted(list(self._group_columns))
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        if isinstance(value, list):
            if len(set(value)) is not len(value):
                raise DuplicateValueError()
            self._group_columns = set(map(lambda x: str(x), value))
        else:
            print("Error: group_columns must be a list.")

    def get_knowledgepack(self, uuid):
        """retrieve knowledgepack by uuid from the server

        Args:
            uuid (str): unique identifier for knowledgepack

        Returns:
            TYPE: a knowledgepack object
        """
        return self._sandbox.knowledgepack(uuid)

    def list_knowledgepacks(self):
        """Lists all of the projects on kb cloud associated with current pipeline

        Returns:
            DataFrame: projects on kb cloud
        """

        knowledgepacks = self._sandbox.get_knowledgepacks().rename(
            columns={
                "name": "Name",
                "project_name": "Project",
                "sandbox_name": "Pipeline",
                "uuid": "kp_uuid",
                "created_at": "Created",
                "knowledgepack_description": "kp_description",
            }
        )
        if len(knowledgepacks) < 1:
            print("No Knowledgepacks stored for this pipeline on the cloud.")
            return None

        return knowledgepacks[
            ["Name", "Created", "Project", "Pipeline", "kp_uuid", "kp_description"]
        ]

    def set_columns(self, data_columns=None, group_columns=None, label_column=None):
        """Sets the columns for group_columns, data_columns and the label column
         to be used in the pipeline. This will automatically handle label column, ignore columns, group columns
         and passthrough columns for the majority of pipelines. For pipelines that need individually specified
         column attributes, set them in the step command.

        Args:
            data_columns (None, list): List of sensor data streams to use.
            group_columns (None, list): List of columns to use when applying aggregate functions
                and defining unique subsets on which to operate.
            label_column (None, str): The column name containing the ground truth label.

        """
        if data_columns:
            self.data_columns = data_columns

        if label_column:
            self.label_column = label_column

        if group_columns:
            self.group_columns = group_columns

    def get_pipeline_length(self):
        """
        Returns:
            int: The current length of the pipeline.
        """
        return len(self._sandbox._pipeline._steps)

    def get_function_type(self, name):
        """
        Returns:
            str: The type of a function.
        """
        return self._kb.functions.get_function_by_name(name).type

    def describe(self, show_params=True, show_set_params=False):
        """Prints out a description of the pipeline steps and parameters

        Args:
            show_params (bool, True): Include the parameters in the pipeline description

        """
        self._sandbox.pipeline.describe(
            show_params=show_params, show_set_params=show_set_params
        )

    def rehydrate_knowledgepack(self, model=None, uuid=None, replace=True):
        """Replace the executing cell with pipeline code for a knowledge Pack

        Args:
            model (model, knowledgepack, None): pass in a model to build a pipeline from that
        """

        return self.rehydrate(model=model, replace=replace, kp_summary=True, uuid=uuid)

    def rehydrate_pipeline(self, model=None, uuid=None, replace=True):
        """Replace the executing cell with pipeline code for the current pipeline or
        pipeline that generated the model

        Args:
            model (model, knowledgepack, None): pass in a model to build a pipeline from that
        """

        return self.rehydrate(model=model, replace=replace, kp_summary=False, uuid=uuid)

    def rehydrate(self, model=None, replace=True, kp_summary=False, uuid=None):
        """Replace the executing cell with pipeline code for either a model or
        pipeline.

        Args:
            model (model, knowledgepack, None): pass in a model to build a pipeline from that
            replace (boolean, True): replace the executing cell with pipeline code
        """
        if isinstance(uuid, str):
            knowledgepack = self._kb.get_knowledgepack(uuid)

        if hasattr(model, "knowledgepack"):
            knowledgepack = model.knowledgepack
        else:
            knowledgepack = model

        steps = None

        if knowledgepack:
            try:
                if kp_summary:
                    if not knowledgepack.knowledgepack_summary:
                        print("Knowledpack doesn't have a summary.")
                        return
                    steps = (
                        [knowledgepack.pipeline_summary[0]]
                        + knowledgepack.knowledgepack_summary["recognition_pipeline"]
                        + [knowledgepack.pipeline_summary[-1]]
                    )
                else:
                    steps = knowledgepack.pipeline_summary
            except:
                raise InvalidModelException(
                    "Model Pipeline was not able to be generated."
                )

        else:
            steps = self._sandbox.pipeline.to_list()

        # reindex min max scale
        if knowledgepack and kp_summary:
            min_max_step = None
            for index in range(len(steps) - 1, 0, -1):
                if steps[index]["name"] == "Min Max Scale":
                    min_max_step = index

            # Set the Min Max Scale features correctly

            feature_min_max = steps[min_max_step]["inputs"][
                "feature_min_max_parameters"
            ]
            rescaled_features_min_max = {"maximums": {}, "minimums": {}}

            for factor in ["maximums", "minimums"]:
                generator_index = 0
                last_index = None
                for _, key in enumerate(sorted(feature_min_max["maximums"].keys())):
                    if last_index != key.split("_")[1]:
                        last_index = key.split("_")[1]
                        generator_index += 1

                    rescaled_features_min_max[factor][
                        "gen_"
                        + "{0:04}".format(generator_index)
                        + "_{}".format("_".join(key.split("_")[2:]))
                    ] = feature_min_max[factor][key]

            steps[min_max_step]["inputs"][
                "feature_min_max_parameters"
            ] = rescaled_features_min_max

            # Add a custom feature selector to handle family generators such as histogram

            fs_dict = generate_custom_feature_selector(knowledgepack.feature_summary)

            fs_step = {
                "type": "selectorset",
                "name": "selector_set",
                "set": [
                    {
                        "inputs": {"custom_feature_selection": fs_dict},
                        "function_name": "Custom Feature Selection By Index",
                    }
                ],
                "inputs": {
                    "remove_columns": [],
                    "passthrough_columns": steps[min_max_step]["inputs"][
                        "passthrough_columns"
                    ],
                    "input_data": steps[min_max_step]["inputs"]["input_data"],
                    "label_column": steps[-1]["label_column"],
                    "number_of_features": len(knowledgepack.feature_summary),
                    "feature_table": steps[min_max_step]["outputs"][1],
                    "cost_function": "sum",
                },
                "outputs": ["temp.selector_set0", "temp.features.selector_set0"],
                "refinement": {},
            }

            steps.insert(min_max_step, fs_step)

        return generate_pipeline(
            self._kb.functions.function_list, steps, replace=replace
        )

    def submit(self, lock=False):
        """Submit a pipeline asynchronously to SensiML Cloud for execution."""
        try:
            self._sandbox.submit()

        except HTTPError:
            return False

        return True

    def execute(self, wait_time=15, silent=True, describe=True, **kwargs):
        """Execute pipeline asynchronously and check for results.

        Args:
            wait_time (int, 10): Time to wait in between status checks.
            silent (bool, True): Silence status updates.
        """
        if describe:
            print("Executing Pipeline with Steps:\n")
            self.describe(show_params=False)

        status = self.submit()

        time.sleep(5)
        return self.get_results(lock=True, wait_time=wait_time, silent=silent, **kwargs)

    def auto(
        self, auto_params, run_parallel=True, lock=True, silent=True, renderer=None
    ):
        """Execute automated pipeline asynchronously.

        The automated pipeline is used to find optimal parameters with a genetic algorithm. The genetic algorithm starts with a user-defined
        randomized population (pipelines) and generates models from them and tests them, keeps a subset of high-performing combinations, and
        then recombines those "survivors" and repeats the process over again. The offspring of good parameter combinations are usually also
        good and sometimes are significantly better than their parents. As the algorithm iterates each successive generation, it often finds a
        near-optimal model without trying as many configurations. The algorithm terminates when the desired number of iteration is completed.

        The created pipelines are run in parallel in SensiML servers and results are ranked by the fitness score which takes into account the
        model's F1 score, precision, sensitivity, and other metrics. The automation options and definition of the performance metrics are
        given below.

        Args:
            seed (str): Pipeline templates that focus on different feature libraries. Options and definitions are given below.
                Basic Features: Generates a set of all-purpose, high-performance features using statistical, energy, and rate of change feature
                    generators. The seed then performs feature selection and model generation algorithms with a genetic algorithm to optimize
                    pipeline parameters.
                Advanced Features: Generates a comprehensive set of features using statistical, energy, amplitude, shape, time, and rate of
                    change feature generators. The seed then performs feature selection and model generation algorithms with a genetic algorithm
                    to optimize pipeline parameters.
                Histogram_Features: Generates a set of histogram features and then performs feature selection and model generation algorithms
                    with a genetic algorithm to optimize pipeline parameters.
                Downsampled_Features: Generates a set of downsampled features and then performs feature selection and model generation algorithms
                    with a genetic algorithm to optimize pipeline parameters.
                Custom: Uses the user-defined pipeline to extract features and then searches for optimal parameters for the feature selection step,
                    if defined, and the model generation step (required) using a genetic algorithm.

            params (dict) : Parameters of the genetic algorithm to optimize the pipelines. Definition and options of the parameters are given below.
                search_steps, (list, ['selectorset', 'tvo']): it is used to define which libraries in the pipelines will be optimized.
                population_size, (int, 10): Initial number of randomly created pipelines.
                iterations, (int, 1): Repetition number of optimization process
                mutation_rate, (float, 0.1): Random changes from the previous population.
                recreation_rate, (float, 0.1): Rate of randomly created pipelines for next generation.
                survivor_rate, (float, 0.5): Ratio of the population that will be transferred to next generation.
                number_of_models_to_return, (int, 5): Number of pipeline that will return to user.
                allow_unknown, (bool, False): Allows creating unknown prediction results for the vectors. A vector is classified as an unknown if it
                    cannot be recognized by any neurons.
                demean_segments, (bool, False): Removes the mean from the input data before extracting features.
                validation_method, (Stratified Shuffle Split): Validation method that will be used in optimization
                balance_data, (bool, False): Use Undersampling of the Majority classes to balance the data prior to Model Building
                outlier_filter, (bool, False): Filter outliers using Isolation Forest Filter with a filter percent of 5.
                fitness (dict): Fitness parameters are combination of statistical and cost variables. It is used to evaluate the performance of
                    the models found by the algorithm. The user will define the coefficient scores for the parameters to define the priority of
                    the parameters in the fitness score. Definition and options of the parameters are given below.

                    statistical variables:
                        accuracy: The degree of correctness of all vectors
                        f1_score: Measures of the test's accuracy
                        precision: Proportion of positive identifications that is actually correct
                        sensitivity: Measures of the proportion of actual positives that are correctly identified
                        specificity: Measures of the proportion of actual negatives that are correctly identified
                        positive_predictive_rate: Ratio of "true positive" is the event that the test makes a positive prediction
                    cost variables:
                        neurons: Number of neurons that used in the model
                        features: Number of features that used in the model,

            lock (bool, False): Ping for results every 30 seconds until the process finishes.
            silent (bool, True): Silence status updates.

        Example:

            activity_data = dsk.datasets.load_activity_raw()
            dsk.upload_dataframe('activity_data.csv', activity_data, force=True)

            dsk.pipeline.reset()
            dsk.pipeline.set_input_data('activity_data.csv',
                                data_columns = ['accelx', 'accely', 'accelz', 'gyrox', 'gyroy', 'gyroz'],
                                group_columns = ['Subject','Class', 'Rep'],
                                label_column = 'Class')

            dsk.pipeline.add_transform("Windowing", params={"window_size": 100, "delta": 100 })

            results, summary = dsk.pipeline.auto({'seed': 'Basic Features',
                                                  'params':{"search_steps": ['selectorset', 'tvo'],
                                                            "population_size": 10,
                                                            "iterations": 1,
                                                            "mutation_rate": 0.1,
                                                            "recreation_rate": 0.1,
                                                            "survivor_rate": 0.5,
                                                            "number_of_models_to_return": 5,
                                                            "run_parallel": True,
                                                            "allow_unknown": False,
                                                            "validation_method": "'Stratified Shuffle Split'",
                                                            "fitness": {'neurons': 0.5,
                                                                        'accuracy': 1.0,
                                                                        'f1_score': 0.0,
                                                                        'features': 0.3,
                                                                        'precision': 0.0,
                                                                        'sensitivity': 0.7,
                                                                        'specificity': 0.0,
                                                                        'positive_predictive_rate': 0.0},

                                                              }})

            results.summarize()



        """
        if self._group_columns:
            auto_params["group_columns"] = list(self._group_columns)
        if self._data_columns:
            auto_params["data_columns"] = list(self._data_columns)
        if self._label_column:
            auto_params["label_column"] = self._label_column

        seed = self._kb.seeds.get_seed_by_name(auto_params["seed"])
        success = self._sandbox.auto(
            auto_params=auto_params, seed=seed, renderer=renderer
        )
        if success:
            return self.get_results(lock=lock, silent=silent, renderer=renderer)
        return (None, None)

    def grid_search(self, grid_params, run_parallel=True, lock=True, silent=True):
        """Grid search is a parameter optimization method that exhaustivley searches over a gridded parameter space.
        Grid search returns will return the score each parameter combination for f1, precision and sensitivity so that
        you can choose the best performing combination to build a knowledge pack with.

        Args:
            grid_params (dict): Grid search parameters.
            run_parallel (bool, True): Run grid search in parallel in KB cloud.
            lock (bool, False): Ping for results every 30 seconds until the process finishes.


        grid_params is a nested python dictionary object.

            grid_params = {"Name Of Function":{"Name of Parameter":[ A, B, C]}}

        Where A, B and C are the parameters to search over. Additionally, for each step you may want to search over more than
        one of a functions configurable parameters. To do this simply add another element to the functions dictionary.

            grid_params = {"Name Of Function":{"Name of Parameter 1":[ A, B, C],
                                            "Name of Parameter 2":[ D, E]}}

        This will tell grid search to search over 6 different parameter spaces.

        You can also specify more than one step to search over in grid params. This is done by adding another element to
        the function level of the grid_params dictionary.

            grid_params = {"Name Of Function":{"Name of Parameter 1":[ A, B, C],
                                            "Name of Parameter 2":[ D, E]},
                          "Name of Function 2":{"Name of Parameter":[1, 2, 3, 4, 5, 6]}}

        Example:

            grid_params = {'Windowing':{"window_size": [100,200],'delta':[100]},
                        'selector_set': {"Recursive Feature Elimination":{'number_of_features':[10, 20]}},
                        'Hierarchical Clustering with Neuron Optimization': {'number_of_neurons':[10,20]}
                        }

            results, stats = dsk.pipeline.grid_search(grid_params)

        """

        print("Executing Pipeline with Steps:\n")
        self.describe(show_params=False)

        self._sandbox.grid_search(grid_params=grid_params, run_parallel=run_parallel)

        time.sleep(5)
        return self.get_results(lock=lock, silent=silent)

    def autosegment_search(self, params, run_parallel=True, lock=True, silent=True):
        """Execute auto segment search pipeline asynchronously.

        Args:
            params (dict): Automation parameters for segment search.
            run_parallel (bool, True): Run in parallel in KB cloud.
            lock (bool, False): Ping for results every 30 seconds until the process finishes.
            silent (bool, True): Silence status updates.
        """
        print("Running AutoSegment Pipeline\n")

        # check that the pipeline is correct for autosegmentation.
        for step_index, step in enumerate(self._sandbox.pipeline.to_list()):
            if step_index == 0:
                if step["type"] != "featurefile":
                    print("First step in pipeline must be a featurefile")
                    return None, None
            else:
                if step["type"] != "transform":
                    print("Pipeline can only contain sensor transforms.")
                    return None, None

        self._sandbox.autosegment_search(params, run_parallel=run_parallel)

        return self.get_results(lock=lock, silent=silent)

    def get_results(
        self,
        lock=False,
        wait_time=15,
        silent=False,
        page_index=0,
        renderer=None,
        **kwargs
    ):
        """Retrieve status, results from the kb cloud for the current pipeline.

        Args:
            results to retrieve. The default is the last type that was run.
            lock (bool, False): This will lock the process and continuously ping the KB cloud
            for the status of the pipeline process.
            wait_time (int, 30): The time to wait between individual status checks.
            silent (bool,  False): This will silence updates to every 4th update check.
            page_index (int, 0): The page desired if the result is multi-paged (1-based)

        Returns:
            results (DataFrame or model result): This is the result of the last executed pipeline step.
            stats (dictionary): A dictionary containing the execution summary, features and other
            summary statistics
        """

        self._results = utility.wait_for_pipeline_result(
            self._sandbox,
            lock=lock,
            wait_time=wait_time,
            silent=silent,
            page_index=page_index,
            renderer=renderer,
            **kwargs
        )

        return self._results[0], self._results[1]

    def data(self, pipeline_step, page_index=0):
        """Retrieves results from a specific pipeline step in the pipeline from stored values in kbcloud
        after execution has been performed.

        Args:
            pipeline_step (int): Pipeline step to retrieve results from.
            page_index (int): Index of data to get.

        Returns:
            A ModelResultSet if the selected pipeline step is TVO step, otherwise the output of the pipeline
            step is returned as a DataFrame.
        """
        try:
            return self._sandbox.intermediate_data(
                pipeline_step=pipeline_step, page_index=page_index
            )[0]
        except HTTPError:
            return None

    def visualize_features(self, feature_vector, label_column=None):
        """Makes a plot of feature vectors by class to aid in understanding your model

        Args:
            feature_vector (DataFrame): Dataframe containing feature vectors and label column
        """
        if label_column is None:
            label_column = self.label_column

        return Visualize(
            feature_vector=feature_vector, label=label_column
        ).plot_features()

    def visualize_neuron_array(
        self, model, feature_vector, featureX, featureY, neuron_alpha=0.2
    ):
        """Makes a plot of feature vectors by class to aid in understanding your model

        Args:
            model (model/knowledpack): The model or knowledpack to use for plotting the neurons
            feature_vector (DataFrame): Dataframe containing feature vectors and label column
            featureX (str): The name of the feature for the x axis
            featureY (str): The name of the feature for the y axis

        """

        if hasattr(model, "knowledgepack"):
            if (
                model.knowledgepack.device_configuration.get("classifier", None)
                != "PME"
            ):
                print(
                    "Visualize neuron array is only supported for models using the PME classifier."
                )
                return

        Visualize(
            model=model, feature_vector=feature_vector, label=self.label_column
        ).neuron_feature_map(featureX, featureY, neuron_alpha=neuron_alpha)

    def clear_cache(self):
        """Deletes the cache on KB cloud for this pipeline."""
        if self._sandbox is not None:
            self._sandbox.delete_cache()

    def reset(self, delete_cache=False):
        """ Reset the current pipeline steps.

        Args:
            delete_cache (bool, False): Delete the cache from KB cloud.
        """
        self._sandbox.clear()
        self._sandbox.update()
        self._generator_index = 0
        self._selector_index = 0
        self._data_columns = ""
        self._tvo_call = None
        self._classifier_call = None
        self._training_algorithm_call = None
        self._validation_call = None
        self._group_columns = None
        self._label_column = None

        if delete_cache:
            print(
                "\n\nWarning:: You have cache set to delete, this will cause your pipelines to run slower!\n\n"
            )
            self.clear_cache()

    def set_knowledgepack_platform(self, *args, **kwargs):
        """Backwards compatible call to set_device_configuration"""
        logger.warning("Deprecated: Please use set_device_configuration instead.")
        self.set_device_configuration(*args, **kwargs)

    def stop_pipeline(self):
        """Kills a pipeline that is running on KB cloud."""
        self._sandbox.kill_pipeline()

    def delete_sandbox(self):
        """Clears the local pipeline steps, and delete the sandbox from the KB cloud."""
        self.reset(delete_cache=True)

        if self._sandbox is not None:
            self._sandbox.delete()

    def _preprocess(self, segmenterid):
        """ adds the segmenter preprocess steps to the pipeline """

        try:
            segmenter = self._project.get_segmenters().loc[segmenterid]
        except TypeError:
            print("Error. Segmenter associated with this query does not exist.")
            return

        if segmenter.preprocess is None:
            return

        preprocess = json.loads(segmenter.preprocess)

        for i in range(len(preprocess.keys())):
            transform = preprocess[str(i)]
            self._add_transform(
                {
                    "name": transform["params"]["name"],
                    "params": transform["params"]["inputs"],
                },
                False,
            )
            self._data_columns.add(transform["actual_name"])
            print("Adding Preprocess Transform {}".format(transform["params"]["name"]))

    def set_input_query(self, name):
        """Set the input data to be a stored query.

        Args:
            name (str): The name of the saved query.
        """

        query_call = self._kb.functions.create_query_call(name)
        query_call.query = self._project.queries.get_query_by_name(name)

        self._dataset = query_call

        label_column = query_call.query.label_column

        self.set_columns(
            query_call.query.columns._columns,
            query_call.query.metadata_columns._columns + [label_column, "SegmentID"],
            label_column,
        )

        self._add_initial_data()

        self._preprocess(query_call.query.segmenter)

    def set_input_data(
        self, name, data_columns=None, group_columns=None, label_column=None
    ):
        """Use a data file that has been uploaded as your data source.

        Args:
            name (str, list): The name of the data file or list of datafiles in SensiML cloud.
            data_columns (list, required): Array of data streams to use in model building.
            group_columns (list, required): The List of columns to use when applying aggregate functions
                and defining unique subsets on which to operate.
            label_column (str, required): The column with the true classification.

        """

        self.set_columns(data_columns, group_columns, label_column)

        if isinstance(name, list):
            for index, csv_name in enumerate(name):
                if csv_name[-4:] != ".csv":
                    name[index] = "{}.csv".format(csv_name)
        else:
            if name[-4:] != ".csv":
                name = "{}.csv".format(name)

        call = self._kb.functions.create_featurefile_call(name)
        self._dataset = call
        self._dataset.data_columns = self.data_columns
        self._dataset.group_columns = self.group_columns
        self._dataset.label_column = self.label_column

        self._add_initial_data()

    def set_input_capture(self, names):
        """Use a data file that has been uploaded as your data source.

        Args:
            name (str,list): single capture or list of captures file names to use in SensiML cloud.

        """

        self.set_columns(list(self._project.columns()), ["Subject"], None)

        call = self._kb.functions.create_capturefile_call(names)
        self._dataset = call
        self._dataset.data_columns = self.data_columns

        self._add_initial_data()

    def add_linear_step(self, func):
        """Add a step to the pipeline. Automatically tie the previous step and current step.

        Args:
            func (function): A sensiml function method call
        """
        self._sandbox.add_linear_step(func)

    def add_segmenter(self, name, params={}):
        """Add a Segmenter to the pipeline.

        Args:
            name (str): Name of the segmenter method to add.
            params (dict, optional): Dictionary containing the parameters of the transform.

        """

        self._add_transform({"name": name, "params": params})

    def add_transform(self, name, params={}, overwrite=True):
        """Add a Transform to the pipeline.

        Args:
            name (str): Name of the transform method to add.
            params (dict, optional): Dictionary containing the parameters of the transform.
            overwrite (boolean): when adding multiple instances of the same transform, set
                overwrite to False for the additional steps and the first instance will not
                be overwritten (default is True)

        """

        self._add_transform({"name": name, "params": params}, overwrite)

    def add_feature_generator(
        self,
        feature_generators,
        params={},
        function_defaults={},
        return_generator_set=False,
    ):
        """Add a feature generator set to the pipeline.

        Args:
            feature_generators (list): List of feature generators. As names or dictionaries.
            params (dict, {}}): Parameters to apply to the feature generator set.
            function_defaults (dict,{}}): Parameters to apply to all individual feature generators.

        Examples:
            >>> # Add a single feature generator
            >>> dsk.pipeline.add_feature_generator([{'name': 'Downsample', 'params': {'new_length': 5}}, {'name': 'Mean'}],
                                                   function_defaults = {'columns': sensor_columns})

            >>> # Call multiple functions by name when they use the same same function defaults
            >>> dsk.pipeline.add_feature_generator(['Mean', 'Standard Deviation', 'Skewness', 'Kurtosis', '25th Percentile',
                                                   '75th Percentile', '100th Percentile', 'Zero Crossing Rate'],
                                                   function_defaults = {'columns': sensor_columns})

            >>> # Call multiple functions using function defaults
            >>> dsk.pipeline.add_feature_generator([{'name': 'Downsample', 'params': {'new_length': 5}}, {'name': 'Mean'}],
                                                    function_defaults = {'columns': sensor_columns})

            >>> # Call multiple functions by subtype which use different parameters; note all subtypes will take the same inputs
            >>> dsk.pipeline.add_feature_generator([{'subtype_call': 'Area', 'params': {'sample_rate': 100, 'smoothing_factor': 9}},
                                                    {'subtype_call': 'Time', 'params': {'sample_rate': 100}},
                                                    {'subtype_call': 'Rate of Change'},
                                                    {'subtype_call': 'Statistical'},
                                                    {'subtype_call': 'Energy'},
                                                    {'subtype_call': 'Amplitude', 'params': {'smoothing_factor': 9}},
                                                    {'subtype_call': 'Physical', 'params': {'sample_rate': 100}}
                                                    ],
                                                    function_defaults={'columns': sensor_columns},
                                                   )

            >>> # Mix subtype and specify additional feature generators
            >>> dsk.pipeline.add_feature_generator([{'subtype_call': 'Statistical'},
                                                    {'name': 'Downsample', 'params': {'new_length': 5}},
                                                    ],
                                                    function_defaults={'columns': sensor_columns},
                                                    )

            >>> # Call the same feature generators multiple times with different parameters
            >>> dsk.pipeline.add_feature_generator([{'name': 'Downsample', 'params': {'new_length': 5, 'columns': sensor_columns[0]}},
                                                    {'name': 'Downsample', 'params': {'new_length': 12}},
                                                    ],
                                                    function_defaults={'columns': sensor_columns},
                                                    )

        """
        # if we get a list of strings
        feature_generators = copy.deepcopy(feature_generators)
        if isinstance(feature_generators[0], str):
            feature_generators = list(
                map(lambda x: {"name": x, "params": {}}, feature_generators)
            )
            for fg in feature_generators:
                for default_param in list(function_defaults.keys()):
                    if default_param not in list(fg["params"].keys()):
                        fg["params"].update(
                            {default_param: function_defaults.get(default_param)}
                        )

        # if we get a list of dicts
        elif isinstance(feature_generators[0], dict):
            for fg in feature_generators:
                if fg.get("params", None) is None:
                    fg["params"] = {}
                for default_param in list(function_defaults.keys()):
                    if default_param not in list(fg["params"].keys()):
                        fg["params"].update(
                            {default_param: function_defaults.get(default_param)}
                        )

        else:
            print(
                "Feature Generator was improperly specified. See Documentation (dsk.pipeline.add_feature_generator?) for examples."
            )
            raise InvalidParameterException(
                "Feature Generator was improperly specified. See Documentation (dsk.pipeline.add_feature_generator?) for examples."
            )

        return self._add_feature_generator(
            feature_generators, params, return_generator_set=return_generator_set
        )

    def add_feature_selector(self, feature_selectors, params={}):
        """Add a feature selection set to the pipeline.

        Args:
            feature_selectors (List): List of dictionaries containing feature selectors
            params (dict, {}): Parameters of the feature selector set.

        Examples:
            >>> dsk.pipeline.add_feature_selector([{"name":"Recursive Feature Elimination", "params":{"method":"Log R"}}],
            >>>                                    params = {"number_of_features":20})
        """

        self._add_feature_selector(feature_selectors, params)

    def set_validation_method(self, name, params={}):
        """Set the validation method for the tvo step.

        Args:
            name (str): Name of the validation method to use.
            params (dict, optional): Parameters for the validation method.

        """
        self._validation_call = call = self._kb.functions.create_validation_method_call(
            name
        )
        for k, v in params.items():
            setattr(call, k, v)

    def set_classifier(self, name, params={}):
        """Classification method for the TVO step to use.

        Args:
            name (str): Name of the classification method.
            params (dict, optional): Parameters of the classification method.
        """
        self._classifier_call = call = self._kb.functions.create_classifier_call(name)
        for k, v in params.items():
            setattr(call, k, v)

    def set_training_algorithm(self, name, params={}):
        """Training algorithm for the TVO step to use.

        Args:
            name (str): Name of the training algorithm.
            params (dict, optional): Parameters of the training algorithm.
        """
        self._training_algorithm_call = (
            call
        ) = self._kb.functions.create_training_algorithm_call(name)
        for k, v in params.items():
            setattr(call, k, v)

    def set_tvo(self, params={}):
        """Description of the train, validate optimize step, which consists of a training algorithm,
        validation method and classifier.

        Args:
            params (dict, optional): Parameters of the TVO step.

        Example:
            >>> dsk.pipeline.set_validation_method('Stratified K-Fold Cross-Validation', params={'number_of_folds':3})
            >>> dsk.pipeline.set_classifier('PME', params={"classification_mode":'RBF','distance_mode':'L1'})
            >>> dsk.pipeline.set_training_algorithm('Hierarchical Clustering with Neuron Optimization', params = {'number_of_neurons':10})
            >>> dsk.pipeline.set_tvo({'label_column':'Label', 'ignore_columns': ['Subject', 'Rep']})
        """
        self._tvo_call = call = self._kb.functions.create_train_and_validation_call(
            "tvo"
        )
        for k, v in params.items():
            setattr(call, k, v)
        if (
            self._classifier_call
            and self._training_algorithm_call
            and self._validation_call
        ):
            call.add_classifier(self._classifier_call)
            call.add_validation_method(self._validation_call)
            call.add_optimizer(self._training_algorithm_call)
        else:
            raise PipelineOrderException(
                "Set classifier, validation, and training_algorithm methods before calling."
            )

        self._add_tvo(self._tvo_call)

    def iterate_columns(self, fg, n_columns=None, columns=None):
        """ Builds Multiple Feature generators by iterating over the input columns

        Args:
            fg (dict): Single input feature generator folowing the standard format
            n_columns (int): The number of columns to return as a combination.
            columns (list): if None, will use the columns in the input feature generator, otherwise will use this list provided

        fg =  {'name':'Downsample with Min Max Scaling', 'params':{"columns": ['gyrZ','gyrX'] , "new_length": 15}}

        fg_new = dsk.pipeline.iterate_columns(fg, n_columns=1)

        >> print(fg_new)
        >> [{'name': 'Downsample with Min Max Scaling', 'params': {'columns': ['gyrX'], 'new_length': 15}},
            {'name': 'Downsample with Min Max Scaling', 'params': {'columns': ['gyrX'], 'new_length': 15}}]

        """

        if n_columns is None:
            function = self._kb.functions.get_function_by_name(fg["name"])
            n_columns = get_contract_column_type(function.input_contract)

            if n_columns is None or n_columns < 0:
                return [fg]

        if columns is None:
            columns = fg["params"]["columns"]

        column_list = itertools.combinations(columns, n_columns)

        fg_list = []
        tmp_fg = copy.deepcopy(fg)

        for column_combo in column_list:
            tmp_fg["params"]["columns"] = list(column_combo)
            fg_list.append(copy.deepcopy(tmp_fg))

        return fg_list

    def _check_for_input_data(self):
        if len(self._sandbox._pipeline.to_list()) == 0:
            raise PipelineOrderException(
                "You must specify the input data before specifying additional steps"
            )

    def _add_tvo(self, call):
        self._add_label_column(call)
        self._add_ignore_columns(call)

        self.tvo_index = self.get_pipeline_length()
        self._sandbox.add_linear_step(call)

    def _add_transform(self, transform, overwrite=True):

        self._check_for_input_data()

        logger.debug("transform:" + str(transform["name"]))

        if (
            self.get_function_type(transform["name"]) == "Segmenter"
            and self.group_columns
        ):
            if "Cascade" not in transform["name"]:
                self._group_columns.discard("SegmentID")
            self._group_columns.discard("CascadeID")

        call = self._kb.functions.create_function_call(transform["name"])
        # Add params to function call object
        for k, v in transform["params"].items():
            setattr(call, k, v)

        self._set_call_columns(call)
        self._sandbox.add_linear_step(call, overwrite)

        if (
            self.get_function_type(transform["name"]) == "Segmenter"
            and self.group_columns
        ):
            self._group_columns.add("SegmentID")
            if "Cascade" in transform["name"]:
                self._group_columns.add("CascadeID")

        if (
            transform["name"] == "Feature Cascade"
            and self.group_columns
        ):
            self._group_columns.add("CascadeID")

    def _add_feature_generator(
        self,
        feature_generators,
        feature_generator_params={},
        return_generator_set=False,
    ):

        self._check_for_input_data()

        generator_set = self._kb.functions.create_generator_call_set("generator_set")

        logger.debug("feature_generator_set")

        # set any generator set params given
        for k, v in feature_generator_params.items():
            setattr(generator_set, k, v)

        subtype_functions = []
        for fg in feature_generators:
            if fg.get("subtype_call", None):
                logger.debug("generator_subtype_call:" + str(fg["subtype_call"]))
                for name in self._kb.list_functions(
                    "Feature Generator",
                    str(fg["subtype_call"]),
                    kp_functions=True,
                    qgrid=False,
                ).NAME:
                    subtype_functions.append({"name": name, "params": fg["params"]})

        feature_generators.extend(subtype_functions)

        feature_generator_list = []
        for fg in feature_generators:
            if fg.get("name", None):
                feature_generator_list.extend(self.iterate_columns(fg))

        for fg in feature_generator_list:
            logger.debug("generator_name:" + str(fg["name"]))
            call = self._kb.functions.create_function_call(fg["name"])

            for k, v in fg["params"].items():
                setattr(call, k, v)

            generator_set.add_generator_call(call)

        if return_generator_set:
            return generator_set

        self._add_group_columns(generator_set)

        self._generator_index = self.get_pipeline_length()

        self._sandbox.add_linear_step(generator_set)

    def _add_feature_selector(self, feature_selectors, selector_set_params={}):

        self._check_for_input_data()

        selector_set = self._kb.functions.create_selector_call_set("selector_set")

        logger.debug("feature_selector_set")

        for k, v in selector_set_params.items():
            setattr(selector_set, k, v)

        for fs in feature_selectors:
            logger.debug("feature_selector:" + fs["name"])
            call = self._kb.functions.create_function_call(fs["name"])
            for k, v in fs["params"].items():
                setattr(call, k, v)
            selector_set.add_selector_call(call)

        self._add_passthrough_columns(selector_set)
        self._add_label_column(selector_set)

        self._selector_index = self.get_pipeline_length()
        self._sandbox.add_linear_step(selector_set)

    def _add_initial_data(self):
        """Add the raw data call to the pipeline """
        if self.get_pipeline_length() != 0:
            raise PipelineOrderException(
                "Input data must be the first step in the pipeline"
            )

        self._sandbox.add_linear_step(self._dataset)

    def _add_label_column(self, call):
        if self.label_column and hasattr(call, "label_column"):
            if call.label_column is None:
                call.label_column = self.label_column

    def _add_group_columns(self, call):
        if self.group_columns and hasattr(call, "group_columns"):
            if call.group_columns is None:
                call.group_columns = self.group_columns

    def _add_passthrough_columns(self, call):
        if self.group_columns and hasattr(call, "passthrough_columns"):
            if call.passthrough_columns is None:
                call.passthrough_columns = self.group_columns

    def _add_ignore_columns(self, call):
        if (self.label_column and self.group_columns) and hasattr(
            call, "ignore_columns"
        ):
            if call.ignore_columns is None:
                call.ignore_columns = list(
                    set(self.group_columns) - set([self.label_column])
                )
        elif hasattr(call, "ignore_columns") and call.ignore_columns is None:
            call.ignore_columns = []

    def _set_call_columns(self, call):
        self._add_label_column(call)
        self._add_group_columns(call)
        self._add_passthrough_columns(call)
        self._add_ignore_columns(call)

    def features_to_tensor(
        self,
        feature_vectors,
        validate=0.2,
        test=0.0,
        label=None,
        feature_columns=None,
        class_map=None,
    ):

        if label is None:
            label = self._label_column

        if class_map is None:
            class_map = {k: v for v, k in enumerate(feature_vectors[label].unique())}

        num_classes = len(class_map)

        feature_vectors["__CAT_LABEL__"] = feature_vectors[label].apply(
            lambda x: class_map[x]
        )

        if feature_columns is None:
            feature_columns = [
                col for col in feature_vectors.columns if col[:4] == "gen_"
            ]

        resample = feature_vectors.sample(frac=1, axis=0).reset_index(drop=True)
        resample_features = resample[feature_columns]

        y_values_encoded = get_one_hot(resample["__CAT_LABEL__"].values, num_classes)

        y_values_encoded = y_values_encoded.reshape(resample.shape[0], num_classes)

        x_values = resample_features.values[:, :].reshape(
            resample_features.shape[0], resample_features.shape[1]
        )

        SAMPLES = x_values.shape[0]
        # Calculate the indices of each section.
        train = 1.0 - (test + validate)
        TRAIN_SPLIT = int(train * SAMPLES)
        TEST_SPLIT = int(0.2 * SAMPLES + TRAIN_SPLIT)

        # Use np.split to chop our data into three parts.
        # The second argument to np.split is an array of indices where the data will be
        # split. We provide two indices, so the data will be divided into three chunks.
        x_train, x_test, x_validate = np.split(x_values, [TRAIN_SPLIT, TEST_SPLIT])
        y_train, y_test, y_validate = np.split(
            y_values_encoded, [TRAIN_SPLIT, TEST_SPLIT]
        )

        return x_train, x_test, x_validate, y_train, y_test, y_validate, class_map

        if label is None:
            label = self._label_column

        if class_map is None:
            class_map = {k: v for v, k in enumerate(feature_vectors[label].unique())}

        num_classes = len(class_map)

        feature_vectors["__CAT_LABEL__"] = feature_vectors[label].apply(
            lambda x: class_map[x]
        )

        if feature_columns is None:
            feature_columns = [
                col for col in feature_vectors.columns if col[:4] == "gen_"
            ]

        resample = feature_vectors.sample(frac=1, axis=0).reset_index(drop=True)
        resample_features = resample[feature_columns]

        y_values_encoded = get_one_hot(
            resample["__CAT_LABEL__"].values, num_classes
        ).reshape((resample.shape[0], 1), num_classes)
        x_values = resample_features.values[:, :].reshape(
            resample_features.shape[0], resample_features.shape[1], 1
        )

        SAMPLES = x_values.shape[0]
        # Calculate the indices of each section.
        train = 1.0 - (test + validate)
        TRAIN_SPLIT = int(train * SAMPLES)
        TEST_SPLIT = int(0.2 * SAMPLES + TRAIN_SPLIT)

        # Use np.split to chop our data into three parts.
        # The second argument to np.split is an array of indices where the data will be
        # split. We provide two indices, so the data will be divided into three chunks.
        x_train, x_test, x_validate = np.split(x_values, [TRAIN_SPLIT, TEST_SPLIT])
        y_train, y_test, y_validate = np.split(
            y_values_encoded, [TRAIN_SPLIT, TEST_SPLIT]
        )

        return x_train, x_test, x_validate, y_train, y_test, y_validate, class_map


def get_contract_column_type(input_contract):

    column_contract = None

    for elem in input_contract:
        if elem.get("name", None) == "columns":
            column_contract = elem
            break

    if isinstance(column_contract, dict):
        return column_contract.get("num_columns", None)

    return None


def generate_custom_feature_selector(feature_summary):
    """
    takes a feature summary and resturns the Custom Feature Selection by Index feature selection dictionary
    """
    current_gen_index = 1
    last_gen_index = feature_summary[0]["Feature"].split("_")[1]
    feature_selection = {1: []}
    for feature in feature_summary:
        if last_gen_index != feature["Feature"].split("_")[1]:
            current_gen_index += 1
            last_gen_index = feature["Feature"].split("_")[1]
            feature_selection[current_gen_index] = []
        feature_selection[current_gen_index].append(feature["GeneratorFamilyIndex"])

    return feature_selection


def get_one_hot(targets, nb_classes):
    res = np.eye(nb_classes)[np.array(targets).reshape(-1)]
    return res.reshape(list(targets.shape) + [nb_classes])

